'''
Defines classes and methods for calculating immunity
'''

import numpy as np
import sciris as sc
from collections.abc import Iterable
from . import utils as hpu
from . import defaults as hpd
from . import parameters as hppar
from . import interventions as hpi


# %% Immunity methods

def init_immunity(sim, create=True):
    ''' Initialize immunity matrices with all genotypes and vaccines in the sim'''

    # Pull out all of the circulating genotypes for cross-immunity
    ng = sim['n_genotypes']

    # Pull out all the vaccination interventions
    vx_intvs = [x for x in sim['interventions'] if isinstance(x, hpi.BaseVaccination)]
    nv = len(vx_intvs)

    # Dimension for immunity matrix
    ndim = ng + nv

    # If immunity values have been provided, process them
    if sim['immunity'] is None or create:

        # Precompute waning - same for all genotypes
        if sim['use_waning']:
            imm_decay = sc.dcp(sim['imm_decay'])
            if 'half_life' in imm_decay.keys():
                imm_decay['half_life'] /= sim['dt']
            sim['imm_kin'] = precompute_waning(t=sim.tvec, pars=imm_decay)

        sim['immunity_map'] = dict()
        # Firstly, initialize immunity matrix with defaults. These are then overwitten with specific values below
        immunity = np.ones((ng, ng), dtype=hpd.default_float)  # Fill with defaults

        # Next, overwrite these defaults with any known immunity values about specific genotypes
        default_cross_immunity = hppar.get_cross_immunity(cross_imm_med=sim['cross_imm_med'], cross_imm_high=sim['cross_imm_high'])
        for i in range(ng):
            sim['immunity_map'][i] = 'infection'
            label_i = sim['genotype_map'][i]
            for j in range(ng):
                label_j = sim['genotype_map'][j]
                if label_i in default_cross_immunity and label_j in default_cross_immunity:
                    immunity[j][i] = default_cross_immunity[label_j][label_i]

        imm_source = ng
        for vi,vx_intv in enumerate(vx_intvs):
            genotype_pars_df = vx_intv.product.genotype_pars[vx_intv.product.genotype_pars.genotype.isin(sim['genotype_map'].values())] # TODO fix this
            vacc_mapping = [genotype_pars_df[genotype_pars_df.genotype==gtype].rel_imm.values[0] for gtype in sim['genotype_map'].values()]
            vacc_mapping += [1]*(vi+1) # Add on some ones to pad out the matrix
            vacc_mapping = np.reshape(vacc_mapping, (len(immunity)+1, 1)).astype(hpd.default_float) # Reshape
            immunity = np.hstack((immunity, vacc_mapping[0:len(immunity),]))
            immunity = np.vstack((immunity, np.transpose(vacc_mapping)))
            vx_intv.product.imm_source = imm_source
            imm_source += 1

        sim['immunity'] = immunity

    sim['immunity'] = sim['immunity'].astype('float32')
    sim['n_imm_sources'] = ndim

    return


def update_peak_immunity(people, inds, imm_pars, imm_source, offset=None, infection=True):
    '''
        Update immunity level

        This function updates the immunity for individuals when an infection or vaccination occurs.
            - individuals that are infected and already have immunity from a previous vaccination/infection have their immunity level;
            - individuals without prior immunity are assigned an initial level drawn from a distribution. This level
                depends on whether the immunity is from a natural infection or from a vaccination (and if so, on the type of vaccine).

        Args:
            people: A people object
            inds: Array of people indices
            imm_pars: Parameters from which to draw values for quantities like ['imm_init'] - either
                        sim pars (for natural immunity) or vaccine pars
            imm_source: index of either genotype or vaccine where immunity is coming from

        Returns: None
    '''

    if infection:
        # Determine whether individual seroconverts based upon genotype
        genotype_label = imm_pars['genotype_map'][imm_source]
        genotype_pars = imm_pars['genotype_pars'][genotype_label]
        seroconvert_probs = np.full(len(inds), fill_value=genotype_pars.sero_prob)
        is_seroconvert = hpu.binomial_arr(seroconvert_probs)

        # Extract parameters and indices
        has_imm = people.nab_imm[imm_source, inds] > 0
        no_prior_imm_inds = inds[~has_imm]
        prior_imm_inds = inds[has_imm]

        if len(prior_imm_inds):
            new_peak = hpu.sample(**imm_pars['imm_init'], size=len(prior_imm_inds))
            people.peak_imm[imm_source, prior_imm_inds] = np.maximum(new_peak, people.peak_imm[imm_source, prior_imm_inds])

            new_cell_imm = hpu.sample(**imm_pars['cell_imm_init'], size=len(prior_imm_inds))
            people.cell_imm[imm_source, prior_imm_inds] = np.maximum(new_cell_imm, people.cell_imm[imm_source, prior_imm_inds])

        if len(no_prior_imm_inds):
            people.peak_imm[imm_source, no_prior_imm_inds] = is_seroconvert[~has_imm] * hpu.sample(**imm_pars['imm_init'], size=len(no_prior_imm_inds))
            people.cell_imm[imm_source, no_prior_imm_inds] = is_seroconvert[~has_imm] * hpu.sample(**imm_pars['cell_imm_init'], size=len(no_prior_imm_inds))
    else:
        # Vaccination by dose
        dose1_inds = inds[people.doses[inds]==1] # First doses
        dose2_inds = inds[people.doses[inds]==2] # Second doses
        dose3_inds = inds[people.doses[inds]==3] # Third doses
        if imm_pars['doses']>1:
            imm_pars['imm_boost'] = sc.promotetolist(imm_pars['imm_boost'])
        if len(dose1_inds)>0: # Initialize immunity for newly vaccinated people
            people.peak_imm[imm_source, dose1_inds] = hpu.sample(**imm_pars['imm_init'], size=len(dose1_inds))
        if len(dose2_inds) > 0: # Boost immunity for people receiving 2nd dose...
            people.peak_imm[imm_source, dose2_inds] *= imm_pars['imm_boost'][0]
        if len(dose3_inds) > 0:
            people.peak_imm[imm_source, dose3_inds] *= imm_pars['imm_boost'][1]

    base_t = people.t + offset if offset is not None else people.t
    people.t_imm_event[imm_source, inds] = base_t

    return


def check_immunity(people):
    '''
    Calculate people's immunity on this timestep from prior infections.
    As an example, suppose HPV16 and 18 are in the sim, and the cross-immunity matrix is:

        pars['immunity'] = np.array([[1., 0.5],
                                     [0.3, 1.]])

    This indicates that people who've had HPV18 have 50% protection against getting 16, and
    people who've had 16 have 30% protection against getting 18.
    Now suppose we have 3 people, whose immunity levels are

        people.nab_imm = np.array([[0.9, 0.0, 0.0],
                               [0.0, 0.7, 0.0]])

    This indicates that person 1 has a prior HPV16 infection, person 2 has a prior HPV18
    infection, and person 3 has no history of infection.

    In this function, we take the dot product of pars['immunity'] and people.nab_imm to get:

        people.sus_imm = np.array([[0.9 , 0.35, 0.  ],
                                   [0.27, 0.7 , 0.  ]])

    This indicates that the person 1 has high protection against reinfection with HPV16, and
    some (30% of 90%) protection against infection with HPV18, and so on.

    '''
    immunity = people.pars['immunity'] # cross-immunity/own-immunity scalars to be applied to immunity level
    sus_imm = np.dot(immunity,people.nab_imm) # Dot product gives immunity to all genotypes
    people.sus_imm[:] = np.minimum(sus_imm, np.ones_like(sus_imm)) # Don't let this be above 1
    return




#%% Methods for computing waning

def precompute_waning(t, pars=None):
    '''
    Process functional form and parameters into values:

        - 'exp_decay'   : exponential decay. Parameters should be init_val and half_life (half_life can be None/nan)
        - 'linear_decay': linear decay

    A custom function can also be supplied.

    Args:
        length (float): length of array to return, i.e., for how long waning is calculated
        pars (dict): passed to individual immunity functions

    Returns:
        array of length 'length' of values
    '''

    pars = sc.dcp(pars)
    form = pars.pop('form')
    choices = [
        'exp_decay',
    ]

    # Process inputs
    if form is None:
        output = np.ones(len(t), dtype=hpd.default_float)

    elif form == 'exp_decay':
        if pars['half_life'] is None: pars['half_life'] = np.nan
        output = exp_decay(t, **pars)

    elif callable(form):
        output = form(t, **pars)

    else:
        errormsg = f'The selected functional form "{form}" is not implemented; choices are: {sc.strjoin(choices)}'
        raise NotImplementedError(errormsg)

    return output


def exp_decay(t, init_val, half_life):
    '''
    Returns an array of length t with values for the immunity at each time step after recovery
    '''
    decay_rate = np.log(2) / half_life if ~np.isnan(half_life) else 0.
    result = init_val * np.exp(-decay_rate * t, dtype=hpd.default_float)
    return result


def linear_decay(length, init_val, slope):
    ''' Calculate linear decay '''
    result = -slope*np.ones(length)
    result[0] = init_val
    return result


def linear_growth(length, slope):
    ''' Calculate linear growth '''
    return slope*np.ones(length)



'''
Set the defaults across each of the different files.
'''

import numpy as np
import sciris as sc
import pylab as pl
from .settings import options as hpo # To set options

# Specify all externally visible functions this file defines -- other things are available as e.g. hpv.defaults.default_int
__all__ = ['datadir', 'default_float', 'default_int', 'get_default_plots']

# Define paths
datadir = sc.path(sc.thisdir(__file__)) / 'data'

#%% Specify what data types to use

result_float = np.float64 # Always use float64 for results, for simplicity
if hpo.precision == 32:
    default_float = np.float32
    default_int   = np.int32
elif hpo.precision == 64: # pragma: no cover
    default_float = np.float64
    default_int   = np.int64
else:
    raise NotImplementedError(f'Precision must be either 32 bit or 64 bit, not {hpo.precision}')


#%% Define all properties of people

class State(sc.prettyobj):
    def __init__(self, name, dtype, fill_value=0, shape=None, label=None, color=None,
                 totalprefix=None):
        '''
        Args:
            name: name of the result as used in the model (e.g. cin1)
            dtype: datatype
            fill_value: default value for this state upon model initialization
            shape: If not none, set to match a string in `pars` containing the dimensionality e.g., `n_genotypes`)
            label: text used to construct labels for the result for displaying on plots and other outputs
            color: color (used for plotting stocks)
            totalprefix: the prefix used for differentiating by-genotype results from total results. Set to None if the result only appears as a total
        '''
        self.name = name
        self.dtype = dtype
        self.fill_value = fill_value
        self.shape = shape
        self.label = label or name
        self.color = color
        self.totalprefix = totalprefix or ('total_' if shape else '')
        return
    
    @property
    def ndim(self):
        return len(sc.tolist(self.shape))+1 # None -> 1, 'n_genotypes' -> 2, etc.
    

    def new(self, pars, n):
        shape = sc.tolist(self.shape) # e.g. convert 'n_genotypes' to ['n_genotypes']
        shape = [pars[s] for s in shape] # e.g. convert ['n_genotypes'] to [2]
        shape.append(n) # We always want to have shape n
        return np.full(shape, dtype=self.dtype, fill_value=self.fill_value)


class PeopleMeta(sc.prettyobj):
    ''' For storing all the keys relating to a person and people '''

    # (attribute, nrows, dtype, default value)
    # If the default value is None, then the array will not be initialized - this is faster and can
    # be used for variables where the People object explicitly supplies the values e.g. age

    # Set the properties of a person
    person = [
        State('uid',            default_int),           # Int
        State('scale',          default_float,  1.0), # Float
        State('level0',         bool,  True), # "Normal" people
        State('level1',         bool,  False), # "High-resolution" people: e.g. cancer agents
        State('age',            default_float,  np.nan), # Float
        State('sex',            default_float,  np.nan), # Float
        State('debut',          default_float,  np.nan), # Float
        State('doses',          default_int,    0),  # Number of doses of the prophylactic vaccine given per person
        State('txvx_doses',     default_int,    0),  # Number of doses of the therapeutic vaccine given per person
        State('vaccine_source', default_int,    -1), # Index of the prophylactic vaccine that individual received
        State('screens',        default_int,    0),  # Number of screens given per person
        State('cin_treatments', default_int,    0),  # Number of CIN treatments given per person
        State('cancer_treatments', default_int,    0),  # Number of cancer treatments given per person
        State('art_adherence',  default_float, 0, label='adherence on ART', color='#aaa8ff')
    ]

    # Set the states that a person can be in
    # The following three groupings are all mutually exclusive and collectively exhaustive.
    alive_states = [
        # States related to whether or not the person is alive or dead
        State('alive',          bool,   True,   label='Population'),    # Save this as a state so we can record population sizes
        State('dead_cancer',    bool,   False,  label='Cumulative cancer deaths'),   # Dead from cancer
        State('dead_other',     bool,   False,  label='Cumulative deaths from other causes'),   # Dead from all other causes
        State('emigrated',      bool,   False,  label='Emigrated'),  # Emigrated
    ]

    viral_states = [
        # States related to whether virus is present
        State('susceptible',    bool, True,     'n_genotypes', label='Number susceptible', color='#4d771e'),               # Allowable dysp states: no_dysp
        State('infectious',     bool, False,    'n_genotypes', label='Number infectious',  color='#c78f65'),               # Allowable dysp states: no_dysp, cin1, cin2, cin3
        State('inactive',       bool, False,    'n_genotypes', label='Number with inactive infection', color='#9e1149'),   # Allowable dysp states: no_dysp, cancer in at least one genotype
    ]

    cell_states = [
        # States related to the cellular changes present in the cervix.
        State('normal',         bool, True, 'n_genotypes', label='Number with no cellular changes', color='#9e1149'), # Allowable viral states: susceptible, infectious, and inactive
        State('episomal',       bool, False, 'n_genotypes', label='Number with episomal infection', color='#9e1149'), # Allowable viral states: susceptible, infectious, and inactive
        State('transformed',    bool, False, 'n_genotypes', label='Number with transformation', color='#9e1149'), # Allowable viral states: susceptible, infectious, and inactive
        State('cancerous',      bool, False, 'n_genotypes', label='Number with cancer', color='#5f5cd2'),      # Allowable viral states: inactive
    ]

    sev = [
        # Markers of disease severity.
        State('sev', default_float, np.nan, shape='n_genotypes'), # Severity of infection, taking values between 0-1
        State('sev_rate', default_float, np.nan, shape='n_genotypes'), # Individual samples from parameters in a logistic function that maps duration of infection to severity
        State('sev_infl', default_float, np.nan, shape='n_genotypes'), # Individual samples from parameters in a logistic function that maps duration of infection to severity
    ]

    derived_states = [
        # From the viral states, cell states, and severity markers, we derive the following additional states:
        State('infected',   bool, False, 'n_genotypes', label='Number infected', color='#c78f65'), # Union of infectious and inactive. Includes people with cancer, people with latent infections, and people with active infections
        State('abnormal',   bool, False, 'n_genotypes', label='Number with abnormal cells', color='#9e1149'),  # Union of episomal, transformed, and cancerous. Allowable viral states: infectious
        State('latent',     bool, False, 'n_genotypes', label='Number with latent infection', color='#5f5cd2'), # Intersection of normal and inactive.
        State('precin',     bool, False, 'n_genotypes', label='Number with precin', color='#9e1149'), # Defined as those with sev < clinical_cuttoff[0]
        State('cin1',       bool, False, 'n_genotypes', label='Number with cin1', color='#9e1149'), # Defined as those with clinical_cuttoff[0] < sev < clinical_cuttoff[1]
        State('cin2',       bool, False, 'n_genotypes', label='Number with cin2', color='#9e1149'), # Defined as those with clinical_cuttoff[1] < sev < clinical_cuttoff[2]
        State('cin3',       bool, False, 'n_genotypes', label='Number with cin3', color='#5f5cd2'), # Defined as those with clinical_cuttoff[2] < sev < clinical_cuttoff[3]
        State('carcinoma',  bool, False, 'n_genotypes', label='Number with carcinoma in situ', color='#5f5cd2'), # Defined as those with clinical_cuttoff[3] < sev < clinical_cuttoff[4]
        State('cin',        bool, False, 'n_genotypes', label='Number with detectable dysplasia', color='#5f5cd2'), # Union of CIN1, CIN3, CIN3, and carcinoma in situ
    ]

    hiv_states = [
        State('hiv',        bool, False, label='Number infected with HIV', color='#5c399c'),
    ]

    # Additional intervention states
    intv_states = [
        State('detected_cancer',    bool,   False, label='Number with detected cancer'), # Whether the person's cancer has been detected
        State('screened',           bool,   False, label='Number screend'), # Whether the person has been screened (how does this change over time?)
        State('cin_treated',        bool,   False, label='Number treated for precancerous lesions'), # Whether the person has been treated for CINs
        State('cancer_treated',     bool,   False, label='Number treated for cancer'), # Whether the person has been treated for cancer
        State('vaccinated',         bool,   False, label='Number vaccinated'), # Whether the person has received the prophylactic vaccine
        State('tx_vaccinated',      bool,   False, label='Number given therapeutic vaccine'), # Whether the person has received the therapeutic vaccine
    ]

    # Collection of mutually exclusive + collectively exhaustive states
    mece_states = alive_states + viral_states + cell_states

    # Collection of states that we store as stock results
    stock_states = viral_states + cell_states + derived_states + intv_states + hiv_states

    # Set dates
    # Convert each MECE state and derived state into a date except for susceptible, alive, and no_dysp (which are True by default)
    dates = [State(f'date_{state.name}', default_float, np.nan, shape=state.shape) for state in mece_states+derived_states+intv_states if not state.fill_value]

    # Immune states, by genotype/vaccine
    imm_states = [
        State('sus_imm',        default_float,  0,'n_imm_sources'),  # Float, by genotype
        State('peak_imm',       default_float,  0,'n_imm_sources'),  # Float, peak level of immunity
        State('nab_imm',            default_float,  0,'n_imm_sources'),  # Float, current immunity level
        State('t_imm_event',    default_int,    0,'n_imm_sources'),  # Int, time since immunity event
        State('cell_imm',       default_float,  0,'n_imm_sources'),
    ]

    # Relationship states
    rship_states = [
        State('rship_start_dates', default_float, np.nan, shape='n_partner_types'),
        State('rship_end_dates', default_float, np.nan, shape='n_partner_types'),
        State('n_rships', default_int, 0, shape='n_partner_types'),
        State('partners', default_float, np.nan, shape='n_partner_types'),  # Int by relationship type
        State('current_partners', default_float, 0, 'n_partner_types'),  # Int by relationship type
    ]

    dates += [
        State('date_clearance',     default_float, np.nan, shape='n_genotypes'),
        State('date_exposed',       default_float, np.nan, shape='n_genotypes'),
    ]

    # Duration of different states: these are floats per person -- used in people.py
    durs = [
        State('dur_infection',      default_float, np.nan, shape='n_genotypes'), # Length of time that a person has any HPV present. Defined for males and females. For females, dur_infection = dur_episomal + dur_transformed. For males, it's taken from a separate distribution
        State('dur_episomal',       default_float, np.nan, shape='n_genotypes'), # Length of time that a person has episomal HPV
        State('dur_transformed',    default_float, np.nan, shape='n_genotypes'), # Length of time that a person has transformed HPV
        State('dur_cancer',         default_float, np.nan, shape='n_genotypes'), # Duration of cancer
    ]

    all_states = person + mece_states + imm_states + hiv_states + intv_states + dates + durs + rship_states + sev

    @classmethod
    def validate(cls):
        """
        Check that states are valid

        This check should be performed when PeopleMeta is consumed (i.e., typically in the People() constructor)
        so that any run-time modifications to the states by the end user get accounted for in validation

        Presently, the only validation check is that the state names are unique, but in principle other
        aspects of the states could be checked too

        :return: None if states are valid
        :raises: ValueError if states are not valid

        """
        # Validate
        state_types = ['person', 'mece_states', 'imm_states', 'hiv_states', 'intv_states', 'dates', 'durs', 'sev', 'all_states']
        for state_type in state_types:
            states = getattr(cls, state_type)
            n_states        = len(states)
            n_unique_states = len(set(states))
            if n_states != n_unique_states: # pragma: no cover
                errormsg = f'In {state_type}, only {n_unique_states} of {n_states} state names are unique'
                raise ValueError(errormsg)

        return


#%% Default result settings

# Flows
# All are stored (1) by genotype and (2) as the total across genotypes
class Flow():
    def __init__(self, name, label=None, color=None, by_genotype=True):
        self.name = name
        self.label = label or name
        self.color = color
        self.by_genotype = by_genotype

flows = [
    Flow('infections',              color='#c78f65',    label='Infections'),
    Flow('dysplasias',              color='#c1ad71',    label='Dysplasias'),
    Flow('precins',                 color='#c1ad71',    label='Pre-CINs'),
    Flow('cin1s',                   color='#c1ad71',    label='CIN1s'),
    Flow('cin2s',                   color='#c1981d',    label='CIN2s'),
    Flow('cin3s',                   color='#b86113',    label='CIN3s'),
    Flow('cancers',                 color='#5f5cd2',    label='Cancers'),
    Flow('detected_cancers',        color='#5f5cd2',    label='Cancer detections', by_genotype=False),
    Flow('cancer_deaths',           color='#000000',    label='Cancer deaths', by_genotype=False),
    Flow('detected_cancer_deaths',  color='#000000',    label='Detected cancer deaths', by_genotype=False),
    Flow('reinfections',            color='#732e26',    label='Reinfections'),
    Flow('reactivations',           color='#732e26',    label='Reactivations'),
    Flow('hiv_infections',          color='#5c399c',    label='HIV infections', by_genotype=False)
]
flow_keys           = [flow.name for flow in flows]
genotype_flow_keys  = [flow.name for flow in flows if flow.by_genotype]

# Stocks: the number in each of the following states
# All are stored (1) by genotype and (2) as the total across genotypes
stock_keys   = [state.name for state in PeopleMeta.stock_states]
stock_names  = [state.label for state in PeopleMeta.stock_states]
stock_colors = [state.color for state in PeopleMeta.stock_states]
total_stock_keys = [state.name for state in PeopleMeta.stock_states if state.shape=='n_genotypes']
other_stock_keys = [state.name for state in PeopleMeta.intv_states+PeopleMeta.hiv_states]

# Incidence. Strong overlap with stocks, but with slightly different naming conventions
# All are stored (1) by genotype and (2) as the total across genotypes
inci_keys   = ['hpv',       'cin1',     'cin2',     'cin3',     'dysplasia',      'cancer']
inci_names  = ['HPV',       'CIN1',     'CIN2',     'CIN3',     'Dysplasia',      'Cancer']
inci_colors = ['#c78f65',   '#c1ad71',  '#c1981d',  '#b86113',  '#c1ad71',  '#5f5cd2']

# Demographics
dem_keys    = ['births',    'other_deaths', 'migration']
dem_names   = ['births',    'other deaths', 'migration']
dem_colors  = ['#fcba03',   '#000000',      '#000000']

# Results by sex
by_sex_keys    = ['infections_by_sex',    'other_deaths_by_sex']
by_sex_names   = ['infections by sex',    'deaths from other causes by sex']
by_sex_colors  = ['#000000',              '#000000']

# Results for storing type distribution by dysplasia
type_dist_keys   = ['precin', 'cin1', 'cin2', 'cin3', 'cancerous']
type_dist_names  = ['Pre-CIN', 'CIN1', 'CIN2', 'CIN3', 'Cancer']

#%% Default initial prevalence

default_init_prev = {
    'age_brackets'  : np.array([  12,   17,   24,   34,  44,   64,    80, 150]),
    'm'             : np.array([ 0.0, 0.05, 0.07, 0.05, 0.02, 0.01, 0.0005, 0]),
    'f'             : np.array([ 0.0, 0.05, 0.07, 0.05, 0.02, 0.01, 0.0005, 0]),
}


#%% Default plotting settings

# Define the 'overview plots', i.e. the most useful set of plots to explore different aspects of a simulation
overview_plots = [
    'infections',
    'dysplasias',
    'cancers',
]

class plot_args():
    ''' Mini class for defining default plot specifications '''
    def __init__(self, keys, name=None, plot_type=None, year=None):
        self.keys = sc.tolist(keys)
        self.name = name
        self.plot_type = plot_type
        self.year = year
        return


def get_default_plots(which='default', kind='sim', sim=None):
    '''
    Specify which quantities to plot; used in sim.py.

    Args:
        which (str): either 'default' or 'overview'
    '''
    which = str(which).lower() # To make comparisons easier

    # Check that kind makes sense
    sim_kind   = 'sim'
    scens_kind = 'scens'
    kindmap = {
        None:      sim_kind,
        'sim':     sim_kind,
        'default': sim_kind,
        'msim':    scens_kind,
        'scen':    scens_kind,
        'scens':   scens_kind,
    }
    if kind not in kindmap.keys():
        errormsg = f'Expecting "sim" or "scens", not "{kind}"'
        raise ValueError(errormsg)
    else:
        is_sim = kindmap[kind] == sim_kind

    # Default plots -- different for sims and scenarios
    if which in ['none', 'default', 'epi']:

        if is_sim:
            plots = sc.objdict({
                'HPV infections by age': 'infections_by_age',
                'HPV prevalence': ['hpv_prevalence_by_genotype'],
                'Pre-cancer prevalence by age': ['precin_prevalence_by_age', 'cin1_prevalence_by_age', 'cin2_prevalence_by_age', 'cin3_prevalence_by_age'],
                'Cancer incidence (per 100,000 women)': ['cancer_incidence', 'asr_cancer_incidence'],
                'Cancers by age': 'cancers_by_age',
                'HPV type distribution': 'type_dist',
            })

        else: # pragma: no cover
            plots = sc.odict({
                'HPV incidence': [
                    'hpv_prevalence',
                ],
                'Cancers per 100,000 women': [
                    'cancer_incidence',
                    ],
            })

    # Demographic plots
    elif which in ['demographic', 'demographics', 'dem', 'demography']:
        if is_sim:
            plots = sc.odict({
                'Birth and death rates': [
                    'cdr',
                    'cbr',
                ],
                'Population size': [
                    'n_alive',
                    'n_alive_by_sex',
                ],
            })

    # Show an overview
    elif which == 'overview': # pragma: no cover
        plots = sc.dcp(overview_plots)

    # Plot absolutely everything
    elif which == 'all': # pragma: no cover
        plots = sim.result_keys('all')

    # Show an overview
    elif 'overview' in which: # pragma: no cover
        plots = sc.dcp(overview_plots)

    else: # pragma: no cover
        errormsg = f'The choice which="{which}" is not supported'
        raise ValueError(errormsg)

    return plots



'''
Specify the core interventions. Other interventions can be
defined by the user by inheriting from these classes.
'''

import numpy as np
import sciris as sc
import pylab as pl
import pandas as pd
import inspect
from . import defaults as hpd
from . import parameters as hppar
from . import utils as hpu
from . import immunity as hpi
from . import base as hpb
from collections import defaultdict

#%% Define data files
datafiles = sc.objdict()
for key in ['dx', 'tx', 'vx']:
    datafiles[key] = hpd.datadir / f'products_{key}.csv'


#%% Helper functions

def find_timepoint(arr, t=None, interv=None, sim=None, which='first'):
    '''
    Helper function to find if the current simulation time matches any timepoint in the
    intervention. Although usually never more than one index is returned, it is
    returned as a list for the sake of easy iteration.

    Args:
        arr (list/function): list of timepoints in the intervention, or a boolean array; or a function that returns these
        t (int): current simulation time (can be None if a boolean array is used)
        interv (intervention): the intervention object (usually self); only used if arr is callable
        sim (sim): the simulation object; only used if arr is callable
        which (str): what to return: 'first', 'last', or 'all' indices

    Returns:
        inds (list): list of matching timepoints; length zero or one unless which is 'all'
    '''
    if callable(arr):
        arr = arr(interv, sim)
        arr = sc.promotetoarray(arr)
    all_inds = sc.findinds(arr=arr, val=t)
    if len(all_inds) == 0 or which == 'all':
        inds = all_inds
    elif which == 'first':
        inds = [all_inds[0]]
    elif which == 'last':
        inds = [all_inds[-1]]
    else: # pragma: no cover
        errormsg = f'Argument "which" must be "first", "last", or "all", not "{which}"'
        raise ValueError(errormsg)
    return inds


def get_subtargets(subtarget, sim):
    '''
    A small helper function to see if subtargeting is a list of indices to use,
    or a function that needs to be called. If a function, it must take a single
    argument, a sim object, and return a list of indices. Also validates the values.
    Currently designed for use with testing interventions, but could be generalized
    to other interventions. Not typically called directly by the user.

    Args:
        subtarget (dict): dict with keys 'inds' and 'vals'; see test_num() for examples of a valid subtarget dictionary
        sim (Sim): the simulation object
    '''

    # Validation
    if callable(subtarget):
        subtarget = subtarget(sim)

    if 'inds' not in subtarget: # pragma: no cover
        errormsg = f'The subtarget dict must have keys "inds" and "vals", but you supplied {subtarget}'
        raise ValueError(errormsg)

    # Handle the two options of type
    if callable(subtarget['inds']): # A function has been provided
        subtarget_inds = subtarget['inds'](sim) # Call the function to get the indices
    else:
        subtarget_inds = subtarget['inds'] # The indices are supplied directly

    # Validate the values
    if callable(subtarget['vals']): # A function has been provided
        subtarget_vals = subtarget['vals'](sim) # Call the function to get the indices
    else:
        subtarget_vals = subtarget['vals'] # The indices are supplied directly
    if sc.isiterable(subtarget_vals):
        if len(subtarget_vals) != len(subtarget_inds): # pragma: no cover
            errormsg = f'Length of subtargeting indices ({len(subtarget_inds)}) does not match length of values ({len(subtarget_vals)})'
            raise ValueError(errormsg)

    return subtarget_inds, subtarget_vals


def select_people(inds, prob=None):
    '''
    Return an array of indices of people to who accept a service being offered

    Args:
        inds: array of indices of people offered a service (e.g. screening, triage, treatment)
        prob: acceptance probability

    Returns: Array of indices of people who accept
    '''
    accept_probs    = np.full_like(inds, fill_value=prob, dtype=hpd.default_float)
    accept_inds     = hpu.true(hpu.binomial_arr(accept_probs))
    return inds[accept_inds]


#%% Generic intervention classes

__all__ = ['Intervention']

class Intervention:
    '''
    Base class for interventions.

    Args:
        label       (str): a label for the intervention (used for plotting, and for ease of identification)
        show_label (bool): whether or not to include the label in the legend
        do_plot    (bool): whether or not to plot the intervention
        line_args  (dict): arguments passed to pl.axvline() when plotting
    '''
    def __init__(self, label=None, show_label=False, do_plot=None, line_args=None, **kwargs):
        # super().__init__(**kwargs)
        self._store_args() # Store the input arguments so the intervention can be recreated
        if label is None: label = self.__class__.__name__ # Use the class name if no label is supplied
        self.label = label # e.g. "Screen"
        self.show_label = show_label # Do not show the label by default
        self.do_plot = do_plot if do_plot is not None else False # Plot the intervention, including if None
        self.line_args = sc.mergedicts(dict(linestyle='--', c='#aaa', lw=1.0), line_args) # Do not set alpha by default due to the issue of overlapping interventions
        self.timepoints = [] # The start and end timepoints of the intervention
        self.initialized = False # Whether or not it has been initialized
        self.finalized = False # Whether or not it has been initialized
        return


    def __repr__(self, jsonify=False):
        ''' Return a JSON-friendly output if possible, else revert to short repr '''

        if self.__class__.__name__ in __all__ or jsonify:
            try:
                json = self.to_json()
                which = json['which']
                pars = json['pars']
                parstr = ', '.join([f'{k}={v}' for k,v in pars.items()])
                output = f"hpv.{which}({parstr})"
            except Exception as E:
                output = f'{type(self)} (error: {str(E)})' # If that fails, print why
            return output
        else:
            return f'{self.__module__}.{self.__class__.__name__}()'


    def __call__(self, *args, **kwargs):
        # Makes Intervention(sim) equivalent to Intervention.apply(sim)
        if not self.initialized:  # pragma: no cover
            errormsg = f'Intervention (label={self.label}, {type(self)}) has not been initialized'
            raise RuntimeError(errormsg)
        return self.apply(*args, **kwargs)


    def disp(self):
        ''' Print a detailed representation of the intervention '''
        return sc.pr(self)


    def _store_args(self):
        ''' Store the user-supplied arguments for later use in to_json '''
        f0 = inspect.currentframe() # This "frame", i.e. Intervention.__init__()
        f1 = inspect.getouterframes(f0) # The list of outer frames
        if self.__class__.__init__ is Intervention.__init__:
            parent = f1[1].frame  # parent = f1[2].frame # The parent frame, e.g. change_beta.__init__()
        else:
            parent = f1[2].frame  # parent = f1[2].frame # The parent frame, e.g. change_beta.__init__()
        _,_,_,values = inspect.getargvalues(parent) # Get the values of the arguments
        if values:
            self.input_args = {}
            for key,value in values.items():
                if key == 'kwargs': # Store additional kwargs directly
                    for k2,v2 in value.items(): # pragma: no cover
                        self.input_args[k2] = v2 # These are already a dict
                elif key not in ['self', '__class__']: # Everything else, but skip these
                    self.input_args[key] = value
        return


    def initialize(self, sim=None):
        '''
        Initialize intervention -- this is used to make modifications to the intervention
        that can't be done until after the sim is created.
        '''
        self.initialized = True
        self.finalized = False
        return


    def finalize(self, sim=None):
        '''
        Finalize intervention

        This method is run once as part of ``sim.finalize()`` enabling the intervention to perform any
        final operations after the simulation is complete (e.g. rescaling)
        '''
        if self.finalized: # pragma: no cover
            raise RuntimeError('Intervention already finalized')  # Raise an error because finalizing multiple times has a high probability of producing incorrect results e.g. applying rescale factors twice
        self.finalized = True
        return


    def apply(self, sim):
        '''
        Apply the intervention. This is the core method which each derived intervention
        class must implement. This method gets called at each timestep and can make
        arbitrary changes to the Sim object, as well as storing or modifying the
        state of the intervention.

        Args:
            sim: the Sim instance

        Returns:
            None
        '''
        raise NotImplementedError


    def shrink(self, in_place=False):
        '''
        Remove any excess stored data from the intervention; for use with sim.shrink().

        Args:
            in_place (bool): whether to shrink the intervention (else shrink a copy)
        '''
        if in_place: # pragma: no cover
            return self
        else:
            return sc.dcp(self)


    def plot_intervention(self, sim, ax=None, **kwargs):
        '''
        Plot the intervention

        This can be used to do things like add vertical lines at timepoints when
        interventions take place. Can be disabled by setting self.do_plot=False.

        Note 1: you can modify the plotting style via the ``line_args`` argument when
        creating the intervention.

        Note 2: By default, the intervention is plotted at the timepoints stored in self.timepoints.
        However, if there is a self.plot_timepoints attribute, this will be used instead.

        Args:
            sim: the Sim instance
            ax: the axis instance
            kwargs: passed to ax.axvline()

        Returns:
            None
        '''
        line_args = sc.mergedicts(self.line_args, kwargs)
        if self.do_plot or self.do_plot is None:
            if ax is None:
                ax = pl.gca()
            if hasattr(self, 'plot_timepoints'):
                timepoints = self.plot_timepoints
            else:
                timepoints = self.timepoints
            if sc.isiterable(timepoints):
                label_shown = False # Don't show the label more than once
                for timepoint in timepoints:
                    if sc.isnumber(timepoint):
                        if self.show_label and not label_shown: # Choose whether to include the label in the legend
                            label = self.label
                            label_shown = True
                        else:
                            label = None
                        date = sim.yearvec[timepoint]
                        ax.axvline(date, label=label, **line_args)
        return


    def to_json(self):
        '''
        Return JSON-compatible representation

        Custom classes can't be directly represented in JSON. This method is a
        one-way export to produce a JSON-compatible representation of the
        intervention. In the first instance, the object dict will be returned.
        However, if an intervention itself contains non-standard variables as
        attributes, then its ``to_json`` method will need to handle those.

        Note that simply printing an intervention will usually return a representation
        that can be used to recreate it.

        Returns:
            JSON-serializable representation (typically a dict, but could be anything else)
        '''
        which = self.__class__.__name__
        pars = sc.jsonify(self.input_args)
        output = dict(which=which, pars=pars)
        return output


#%% Template classes for routine and campaign delivery
__all__ += ['RoutineDelivery', 'CampaignDelivery']

class RoutineDelivery(Intervention):
    '''
    Base class for any intervention that uses routine delivery; handles interpolation of input years.
    '''
    def __init__(self, years=None, start_year=None, end_year=None, prob=None, annual_prob=True):
        self.years      = years
        self.start_year = start_year
        self.end_year   = end_year
        self.prob       = sc.promotetoarray(prob)
        self.annual_prob = annual_prob # Determines whether the probability is annual or per timestep
        return

    def initialize(self, sim):

        # Validate inputs
        if (self.years is not None) and (self.start_year is not None or self.end_year is not None):
            errormsg = 'Provide either a list of years or a start year, not both.'
            raise ValueError(errormsg)

        # If start_year and end_year are not provided, figure them out from the provided years or the sim
        if self.years is None:
            if self.start_year is None: self.start_year = sim['start']
            if self.end_year is None:   self.end_year   = sim['end']
        else:
            self.start_year = self.years[0]
            self.end_year   = self.years[-1]

        # More validation
        if (self.start_year not in sim.yearvec) or (self.end_year not in sim.yearvec):
            errormsg = 'Years must be within simulation start and end dates.'
            raise ValueError(errormsg)

        # Adjustment to get the right end point
        adj_factor = int(1/sim['dt'])-1 if sim['dt']<1 else 1

        # Determine the timepoints at which the intervention will be applied
        self.start_point    = sc.findinds(sim.yearvec, self.start_year)[0]
        self.end_point      = sc.findinds(sim.yearvec, self.end_year)[0] + adj_factor
        self.years          = sc.inclusiverange(self.start_year, self.end_year)
        self.timepoints     = sc.inclusiverange(self.start_point, self.end_point)
        self.yearvec        = np.arange(self.start_year, self.end_year+adj_factor, sim['dt'])

        # Get the probability input into a format compatible with timepoints
        if len(self.years) != len(self.prob):
            if len(self.prob)==1:
                self.prob = np.array([self.prob[0]]*len(self.timepoints))
            else:
                errormsg = f'Length of years incompatible with length of probabilities: {len(self.years)} vs {len(self.prob)}'
                raise ValueError(errormsg)
        else:
            self.prob = sc.smoothinterp(self.yearvec, self.years, self.prob, smoothness=0)

        # Lastly, adjust the probability by the sim's timestep, if it's an annual probability
        if self.annual_prob: self.prob = 1-(1-self.prob)**sim['dt']

        return


class CampaignDelivery(Intervention):
    '''
    Base class for any intervention that uses campaign delivery; handles interpolation of input years.
    '''
    def __init__(self, years, interpolate=None, prob=None, annual_prob=True):
        self.years = sc.promotetoarray(years)
        self.interpolate = True if interpolate is None else interpolate
        self.prob = sc.promotetoarray(prob)
        self.annual_prob = annual_prob
        return

    def initialize(self, sim):
        # Decide whether to apply the intervention at every timepoint throughout the year, or just once.
        if self.interpolate:
            self.timepoints = hpu.true(np.isin(np.floor(sim.yearvec), np.floor(self.years)))
        else:
            self.timepoints = hpu.true(np.isin(sim.yearvec, self.years))

        # Get the probability input into a format compatible with timepoints
        if len(self.prob) == len(self.years) and self.interpolate:
            self.prob = sc.smoothinterp(np.arange(len(self.timepoints)), np.arange(len(self.years)), self.prob, smoothness=0)
        elif len(self.prob) == 1:
            self.prob = np.array([self.prob[0]] * len(self.timepoints))
        else:
            errormsg = f'Length of years incompatible with length of probabilities: {len(self.years)} vs {len(self.prob)}'
            raise ValueError(errormsg)

        # Lastly, adjust the annual probability by the sim's timestep, if it's an annual probability
        if self.annual_prob: self.prob = 1-(1-self.prob)**sim['dt']

        return


#%% Behavior change interventions
__all__ += ['dynamic_pars', 'EventSchedule', 'set_intervention_attributes']

class dynamic_pars(Intervention):
    '''
    A generic intervention that modifies a set of parameters at specified points
    in time.

    The intervention takes a single argument, pars, which is a dictionary of which
    parameters to change, with following structure: keys are the parameters to change,
    then subkeys 'days' and 'vals' are either a scalar or list of when the change(s)
    should take effect and what the new value should be, respectively.

    You can also pass parameters to change directly as keyword arguments.

    Args:
        pars (dict): described above
        kwargs (dict): passed to Intervention()

    **Examples**::

        interv = hpv.dynamic_pars(condoms=dict(timepoints=10, vals={'c':0.9})) # Increase condom use amount casual partners to 90%
        interv = hpv.dynamic_pars({'beta':{'timepoints':[10, 15], 'vals':[0.005, 0.015]}, # At timepoint 10, reduce beta, then increase it again
                                  'debut':{'timepoints':10, 'vals':dict(f=dict(dist='normal', par1=20, par2=2.1), m=dict(dist='normal', par1=19.6, par2=1.8))}}) # Increase mean age of sexual debut
    '''

    def __init__(self, pars=None, **kwargs):

        # Find valid sim parameters and move matching keyword arguments to the pars dict
        pars = sc.mergedicts(pars) # Ensure it's a dictionary
        sim_par_keys = list(hppar.make_pars().keys()) # Get valid sim parameters
        kwarg_keys = [k for k in kwargs.keys() if k in sim_par_keys]
        for kkey in kwarg_keys:
            pars[kkey] = kwargs.pop(kkey)

        # Do standard initialization
        super().__init__(**kwargs) # Initialize the Intervention object

        # Handle the rest of the initialization
        subkeys = ['timepoints', 'vals']
        for parkey in pars.keys():
            for subkey in subkeys:
                if subkey not in pars[parkey].keys(): # pragma: no cover
                    errormsg = f'Parameter {parkey} is missing subkey {subkey}'
                    raise sc.KeyNotFoundError(errormsg)
                if sc.isnumber(pars[parkey][subkey]):
                    pars[parkey][subkey] = sc.promotetoarray(pars[parkey][subkey])
                else:
                    pars[parkey][subkey] = sc.promotetolist(pars[parkey][subkey])
        self.pars = pars

        return

    def initialize(self, sim):
        ''' Initialize with a sim '''
        for parkey in self.pars.keys():
            try: # First try to interpret the timepoints as dates
                tps = sim.get_t(self.pars[parkey]['timepoints'])  # Translate input to timepoints
            except:
                tps = []
                # See if it's in the time vector
                for tp in self.pars[parkey]['timepoints']:
                    if tp in sim.tvec:
                        tps.append(tp)
                    else: # Give up
                        errormsg = f'Could not parse timepoints provided for {parkey}.'
                        raise ValueError(errormsg)
            self.pars[parkey]['processed_timepoints'] = sc.promotetoarray(tps)
        self.initialized = True
        return


    def apply(self, sim):
        ''' Loop over the parameters, and then loop over the timepoints, applying them if any are found '''
        t = sim.t
        for parkey,parval in self.pars.items():
            if t in parval['processed_timepoints']:
                self.timepoints.append(t)
                ind = sc.findinds(parval['processed_timepoints'], t)[0]
                val = parval['vals'][ind]
                if isinstance(val, dict):
                    sim[parkey].update(val) # Set the parameter if a nested dict
                else:
                    sim[parkey] = val # Set the parameter if not a dict
        return


class EventSchedule(Intervention):
    """
    Run functions on different days

    This intervention is a a kind of generalization of ``dynamic_pars`` to allow more
    flexibility in triggering multiple, arbitrary operations and to more easily assemble
    multiple changes at different times. This intervention can be used to implement scale-up
    or other changes to interventions without needing to implement time-dependency in the
    intervention itself.

    To use the intervention, simply index the intervention by ``t`` or by date.

    Example:

    >>> iv = EventSchedule()
    >>> iv[1] = lambda sim: print(sim.t)
    >>> iv['2020-04-02'] = lambda sim: print('foo')

    """

    def __init__(self):
        super().__init__()
        self.schedule = defaultdict(list)

    def __getitem__(self, day):
        return self.schedule[day]

    def __setitem__(self, day, fcn):
        if day in self.schedule:
            raise Exception("Use a list instead to assign multiple functions - or to really overwrite, delete the function for this day first i.e. del schedule[day] before performing schedule[day]=...")
        self.schedule[day] = fcn

    def __delitem__(self, key):
        del self.schedule[key]

    def initialize(self, sim):
        super().initialize(sim)

        # First convert all values into lists (i.e., wrap any standalone functions into lists)
        for k, v in list(self.schedule.items()):
            self.schedule[k] = [v] if not isinstance(self.schedule[k], list) else v

        # Then convert any dates into time indices
        for k, v in list(self.schedule.items()):
            t = sim.get_t(k)[0]
            if t != k:
                self.schedule[t] += v
                del self.schedule[k]

    def apply(self, sim):
        if sim.t in self.schedule:
            for fcn in self.schedule[sim.t]:
                fcn(sim)


def set_intervention_attributes(sim, intervention_name, **kwargs):
    # This is a helper method that can be used to set arbitrary intervention attributes
    # It's a separately defined function so that it can be pickled properly
    iv = sim.get_intervention(intervention_name)
    for attr, value in kwargs.items():
        assert hasattr(iv, attr), "set_intervention_attributes() should only be used to change existing attributes"  # avoid silent errors if the attr is misspelled
        setattr(iv, attr, value)


#%% Vaccination
__all__ += ['BaseVaccination', 'routine_vx', 'campaign_vx']


class BaseVaccination(Intervention):
    '''
    Base vaccination class for determining who will receive a vaccine.

    Args:
         product        (str/Product)   : the vaccine to use
         prob           (float/arr)     : annual probability of eligible population getting vaccinated
         age_range      (list/tuple)    : age range to vaccinate
         sex            (int/str/list)  : sex to vaccinate - accepts 0/1 or 'f'/'m' or a list of both
         eligibility    (inds/callable) : indices OR callable that returns inds
         label          (str)           : the name of vaccination strategy
         kwargs         (dict)          : passed to Intervention()
    '''
    def __init__(self, product=None, prob=None, age_range=None, sex=None, eligibility=None, label=None, **kwargs):
        Intervention.__init__(self, **kwargs)
        self.prob = sc.promotetoarray(prob)
        self.age_range = age_range
        self.label = label
        self.eligibility = eligibility
        self._parse_product(product)

        # Deal with sex
        if sc.checktype(sex,'listlike'):
            if sc.checktype(sex[0],'str'): # If provided as ['f','m'], convert to [0,1]
                self.sex = np.array([0,1])
        elif sc.checktype(sex, 'str'):  # If provided as 'f' or 'm', convert to 0 or 1
            if sex=='f':
                self.sex = np.array([0])
            elif sex=='m':
                self.sex = np.array([1])
            else:
                errormsg = f'Sex "{sex}" not understood.'
                raise ValueError(errormsg)
        else:
            self.sex = sc.promotetoarray(sex)

    def _parse_product(self, product):
        '''
        Parse the product input
        '''
        if isinstance(product, Product): # No need to do anything
            self.product=product
        elif isinstance(product, str): # Try to find it in the list of defaults
            try:
                self.product = default_vx(prod_name=product)
            except:
                errormsg = f'Could not find product {product} in the standard list.'
                raise ValueError(errormsg)
        else:
            errormsg = f'Cannot understand format of product {product} - please provide it as either a Product or string matching a default product.'
            raise ValueError(errormsg)
        return

    def initialize(self, sim):
        Intervention.initialize(self)
        self.npts = sim.res_npts
        self.n_products_used = hpb.Result(name=f'Products administered by {self.label}', npts=sim.res_npts, scale=True)
        return


    def check_eligibility(self, sim):
        '''
        Determine who is eligible for vaccination
        '''
        conditions = sim.people.alive # Start by assuming everyone alive is eligible
        if len(self.sex)==1:
            conditions = conditions & (sim.people.sex == self.sex[0]) # Filter by sex
        if self.age_range is not None:
            conditions = conditions & ((sim.people.age >= self.age_range[0]) & (sim.people.age < self.age_range[1])) # Filter by age
        if self.eligibility is not None:
            other_eligible  = sc.promotetoarray(self.eligibility(sim)) # Apply any other user-defined eligibility
            conditions      = conditions & other_eligible
        return hpu.true(conditions)


    def apply(self, sim):
        '''
        Perform vaccination by finding who's eligible for vaccination, finding who accepts, and applying the vaccine product.
        '''
        # Determine whether to apply the intervention. Apply it if no timepoints have been given or
        # if the timepoint matches one of the requested timepoints.
        if len(self.timepoints)>0 and (sim.t not in self.timepoints): do_apply = False
        else: do_apply = True
        accept_inds = np.array([])

        if do_apply:

            # Select people for screening and then record the number of screens
            eligible_inds   = self.check_eligibility(sim) # Check eligibility
            if len(self.timepoints)==0: # No timepoints provided
                prob = self.prob
            else: # Get the proportion of people who screen on this timestep
                prob = self.prob[sc.findinds(self.timepoints, sim.t)[0]]
            accept_inds = select_people(eligible_inds, prob=prob)

            if len(accept_inds):
                self.product.administer(sim.people, accept_inds) # Administer the product

                # Update people's state and dates
                sim.people.vaccinated[accept_inds] = True
                sim.people.date_vaccinated[accept_inds] = sim.t
                sim.people.doses[accept_inds] += 1

                # Update results
                idx = int(sim.t / sim.resfreq)
                new_vx_inds = hpu.ifalsei(sim.people.vaccinated, accept_inds)  # Figure out people who are getting vaccinated for the first time
                n_new_doses = sim.people.scale_flows(accept_inds)  # Scale
                n_new_people = sim.people.scale_flows(new_vx_inds)  # Scale
                sim.results['new_vaccinated'][:,idx] += n_new_people
                sim.results['new_doses'][idx] += n_new_doses
                self.n_products_used[idx] += n_new_doses

        return accept_inds


    def shrink(self, in_place=True):
        ''' Shrink vaccination intervention '''
        obj = super().shrink(in_place=in_place)
        obj.vaccinated = None
        return obj


class routine_vx(BaseVaccination, RoutineDelivery):
    '''
    Routine vaccination - an instance of base vaccination combined with routine delivery.
    See base classes for a description of input arguments.

    **Examples**::

        vx1 = hpv.routine_vx(product='bivalent', age_range=[9,10], prob=0.9, start_year=2025) # Vaccinate 90% of girls aged 9-10 every year
        vx2 = hpv.routine_vx(product='bivalent', age_range=[9,10], prob=0.9, sex=[0,1], years=np.arange(2020,2025)) # Screen 90% of girls and boys aged 9-10 every year from 2020-2025
        vx3 = hpv.routine_vx(product='quadrivalent', prob=np.linspace(0.2,0.8,5), years=np.arange(2020,2025)) # Scale up vaccination over 5 years starting in 2020
    '''

    def __init__(self, product=None, prob=None, age_range=None, sex=0, eligibility=None,
                 start_year=None, end_year=None, years=None, **kwargs):

        BaseVaccination.__init__(self, product=product, age_range=age_range, sex=sex, eligibility=eligibility, **kwargs)
        RoutineDelivery.__init__(self, prob=prob, start_year=start_year, end_year=end_year, years=years)

    def initialize(self, sim):
        RoutineDelivery.initialize(self, sim) # Initialize this first, as it ensures that prob is interpolated properly
        BaseVaccination.initialize(self, sim) # Initialize this next


class campaign_vx(BaseVaccination, CampaignDelivery):
    '''
    Campaign vaccination - an instance of base vaccination combined with campaign delivery.
    See base classes for a description of input arguments.
    '''

    def __init__(self, product=None, prob=None, age_range=None, sex=0, eligibility=None,
                 years=None, interpolate=True, **kwargs):

        BaseVaccination.__init__(self, product=product, age_range=age_range, sex=sex, eligibility=eligibility, **kwargs)
        CampaignDelivery.__init__(self, prob=prob, years=years, interpolate=interpolate)

    def initialize(self, sim):
        CampaignDelivery.initialize(self, sim) # Initialize this first, as it ensures that prob is interpolated properly
        BaseVaccination.initialize(self, sim) # Initialize this next


#%% Screening and triage
__all__ += ['BaseTest', 'BaseScreening', 'routine_screening', 'campaign_screening', 'BaseTriage', 'routine_triage', 'campaign_triage']


class BaseTest(Intervention):
    '''
    Base class for screening and triage.

    Args:
         product        (str/Product)   : the diagnostic to use
         prob           (float/arr)     : annual probability of eligible women receiving the diagnostic
         eligibility    (inds/callable) : indices OR callable that returns inds
         label          (str)           : the name of screening strategy
         kwargs         (dict)          : passed to Intervention()
    '''

    def __init__(self, product=None, prob=None, eligibility=None, **kwargs):
        Intervention.__init__(self, **kwargs)
        self.prob           = sc.promotetoarray(prob)
        self.eligibility    = eligibility
        self._parse_product(product)

    def _parse_product(self, product):
        '''
        Parse the product input
        '''
        if isinstance(product, Product): # No need to do anything
            self.product=product
        elif isinstance(product, str): # Try to find it in the list of defaults
            try:
                self.product = default_dx(prod_name=product)
            except:
                errormsg = f'Could not find product {product} in the standard list.'
                raise ValueError(errormsg)
        else:
            errormsg = f'Cannot understand format of product {product} - please provide it as either a Product or string matching a default product.'
            raise ValueError(errormsg)
        return

    def initialize(self, sim):
        Intervention.initialize(self)
        self.npts = sim.res_npts
        self.n_products_used = hpb.Result(name=f'Products administered by {self.label}', npts=sim.res_npts, scale=True)
        self.outcomes = {k:np.array([], dtype=hpd.default_int) for k in self.product.hierarchy}
        return

    def deliver(self, sim):
        '''
        Deliver the diagnostics by finding who's eligible, finding who accepts, and applying the product.
        '''
        ti = sc.findinds(self.timepoints, sim.t)[0]
        prob = self.prob[ti] # Get the proportion of people who will be tested this timestep
        eligible_inds = self.check_eligibility(sim) # Check eligibility
        accept_inds = select_people(eligible_inds, prob=prob) # Find people who accept
        if len(accept_inds):
            idx = int(sim.t / sim.resfreq)
            self.n_products_used[idx] += sim.people.scale_flows(accept_inds)
            self.outcomes = self.product.administer(sim, accept_inds) # Actually administer the diagnostic, filtering people into outcome categories
        return accept_inds

    def check_eligibility(self, sim):
        raise NotImplementedError


class BaseScreening(BaseTest):
    '''
    Base class for screening.

    Args:
        age_range (list/tuple/arr)  : age range for screening, e.g. [30,50]
        kwargs    (dict)            : passed to BaseTest
    '''
    def __init__(self, age_range=None, **kwargs):
        BaseTest.__init__(self, **kwargs) # Initialize the BaseTest object
        self.age_range = age_range or [30,50] # This is later filtered to exclude people not yet sexually active

    def check_eligibility(self, sim):
        '''
        Return an array of indices of agents eligible for screening at time t, i.e. sexually active
        females in age range, plus any additional user-defined eligibility, which often includes
        the screening interval.
        '''
        adult_females   = sim.people.is_female_adult
        in_age_range    = (sim.people.age >= self.age_range[0]) * (sim.people.age <= self.age_range[1])
        conditions      = (adult_females * in_age_range).astype(bool)
        if self.eligibility is not None:
            other_eligible  = sc.promotetoarray(self.eligibility(sim))
            conditions      = conditions * other_eligible
        return hpu.true(conditions)

    def apply(self, sim):
        '''
        Perform screening by finding who's eligible, finding who accepts, and applying the product.
        '''
        self.outcomes = {k:np.array([], dtype=hpd.default_int) for k in self.product.hierarchy}
        accept_inds = np.array([])
        if sim.t in self.timepoints:
            accept_inds = self.deliver(sim)
            sim.people.screened[accept_inds] = True
            sim.people.screens[accept_inds] += 1
            sim.people.date_screened[accept_inds] = sim.t

            # Store results
            idx = int(sim.t / sim.resfreq)
            new_screen_inds = hpu.ifalsei(sim.people.screened, accept_inds)  # Figure out people who are getting screened for the first time
            n_new_people = sim.people.scale_flows(new_screen_inds)  # Scale
            n_new_screens = sim.people.scale_flows(accept_inds)  # Scale
            sim.results['new_screened'][idx] += n_new_people
            sim.results['new_screens'][idx] += n_new_screens

        return accept_inds


class BaseTriage(BaseTest):
    '''
    Base class for triage.

    Args:
        kwargs (dict): passed to BaseTest
    '''
    def __init__(self, **kwargs):
        BaseTest.__init__(self, **kwargs)

    def check_eligibility(self, sim):
        return sc.promotetoarray(self.eligibility(sim))

    def apply(self, sim):
        self.outcomes = {k:np.array([], dtype=hpd.default_int) for k in self.product.hierarchy}
        accept_inds = np.array([])
        if sim.t in self.timepoints: accept_inds = self.deliver(sim)
        return accept_inds


class routine_screening(BaseScreening, RoutineDelivery):
    '''
    Routine screening - an instance of base screening combined with routine delivery.
    See base classes for a description of input arguments.

    **Examples**::

        screen1 = hpv.routine_screening(product='hpv', prob=0.02) # Screen 2% of the eligible population every year
        screen2 = hpv.routine_screening(product='hpv', prob=0.02, start_year=2020) # Screen 2% every year starting in 2020
        screen3 = hpv.routine_screening(product='hpv', prob=np.linspace(0.005,0.025,5), years=np.arange(2020,2025)) # Scale up screening over 5 years starting in 2020
    '''
    def __init__(self, product=None, prob=None, eligibility=None, age_range=None,
                         years=None, start_year=None, end_year=None, **kwargs):
        BaseScreening.__init__(self, product=product, age_range=age_range, eligibility=eligibility, **kwargs)
        RoutineDelivery.__init__(self, prob=prob, start_year=start_year, end_year=end_year, years=years)

    def initialize(self, sim):
        RoutineDelivery.initialize(self, sim) # Initialize this first, as it ensures that prob is interpolated properly
        BaseScreening.initialize(self, sim) # Initialize this next

class campaign_screening(BaseScreening, CampaignDelivery):
    '''
    Campaign screening - an instance of base screening combined with campaign delivery.
    See base classes for a description of input arguments.

    **Examples**::

        screen1 = hpv.campaign_screening(product='hpv', prob=0.2, years=2030) # Screen 20% of the eligible population in 2020
        screen2 = hpv.campaign_screening(product='hpv', prob=0.02, years=[2025,2030]) # Screen 20% of the eligible population in 2025 and again in 2030
    '''
    def __init__(self, product=None, age_range=None, sex=None, eligibility=None,
                 prob=None, years=None, interpolate=None, **kwargs):
        BaseScreening.__init__(self, product=product, age_range=age_range, sex=sex, eligibility=eligibility, **kwargs)
        CampaignDelivery.__init__(self, prob=prob, years=years, interpolate=interpolate)

    def initialize(self, sim):
        CampaignDelivery.initialize(self, sim) # Initialize this first, as it ensures that prob is interpolated properly
        BaseScreening.initialize(self, sim) # Initialize this next


class routine_triage(BaseTriage, RoutineDelivery):
    '''
    Routine triage - an instance of base triage combined with routine delivery.
    See base classes for a description of input arguments.
    
    **Examples**::

        # Example 1: Triage 40% of the eligible population in all years
        triage1 = hpv.routine_triage(product='via_triage', prob=0.4)

        # Example 2: Triage positive screens into confirmatory testing or theapeutic vaccintion
        screened_pos = lambda sim: sim.get_intervention('screening').outcomes['positive']
        triage2 = hpv.routine_triage(product='pos_screen_assessment', eligibility=screen_pos, prob=0.9, start_year=2030)
    '''
    def __init__(self, product=None, prob=None, eligibility=None, age_range=None,
                         years=None, start_year=None, end_year=None, annual_prob=None, **kwargs):
        BaseTriage.__init__(self, product=product, age_range=age_range, eligibility=eligibility, **kwargs)
        RoutineDelivery.__init__(self, prob=prob, start_year=start_year, end_year=end_year, years=years, annual_prob=annual_prob)

    def initialize(self, sim):
        RoutineDelivery.initialize(self, sim) # Initialize this first, as it ensures that prob is interpolated properly
        BaseTriage.initialize(self, sim) # Initialize this next


class campaign_triage(BaseTriage, CampaignDelivery):
    '''
    Campaign triage - an instance of base triage combined with campaign delivery.
    See base classes for a description of input arguments.
    
    **Examples**::

        # Example 1: In 2030, triage all positive screens into confirmatory testing or therapeutic vaccintion
        screened_pos = lambda sim: sim.get_intervention('screening').outcomes['positive']
        triage1 = hpv.campaign_triage(product='pos_screen_assessment', eligibility=screen_pos, prob=0.9, years=2030)
    '''
    def __init__(self, product=None, age_range=None, sex=None, eligibility=None,
                 prob=None, years=None, interpolate=None, annual_prob=None, **kwargs):
        BaseTriage.__init__(self, product=product, age_range=age_range, sex=sex, eligibility=eligibility, **kwargs)
        CampaignDelivery.__init__(self, prob=prob, years=years, interpolate=interpolate, annual_prob=annual_prob)

    def initialize(self, sim):
        CampaignDelivery.initialize(self, sim) # Initialize this first, as it ensures that prob is interpolated properly
        BaseTriage.initialize(self, sim) # Initialize this next



#%% Treatment interventions
__all__ += ['BaseTreatment', 'treat_num', 'treat_delay', 'BaseTxVx', 'routine_txvx', 'campaign_txvx', 'linked_txvx',]

class BaseTreatment(Intervention):
    '''
    Base treatment class.

    Args:
         product        (str/Product)   : the treatment product to use
         accept_prob     (float/arr)    : acceptance rate of treatment - interpreted as the % of women eligble for treatment who accept
         eligibility    (inds/callable) : indices OR callable that returns inds
         label          (str)           : the name of treatment strategy
         kwargs         (dict)          : passed to Intervention()
    '''
    def __init__(self, product=None, prob=None, eligibility=None, age_range=None, **kwargs):
        Intervention.__init__(self, **kwargs)
        self.prob = sc.promotetoarray(prob)
        self.eligibility = eligibility
        self._parse_product(product)
        self.age_range = age_range or [0,99] # By default, no restrictions on treatment age


    def _parse_product(self, product):
        '''
        Parse the product input
        '''
        if isinstance(product, Product): # No need to do anything
            self.product=product
        elif isinstance(product, str): # Try to find it in the list of defaults
            try:
                self.product = default_tx(prod_name=product)
            except:
                errormsg = f'Could not find product {product} in the standard list.'
                raise ValueError(errormsg)
        else:
            errormsg = f'Cannot understand format of product {product} - please provide it as either a Product or string matching a default product.'
            raise ValueError(errormsg)
        return

    def initialize(self, sim):
        Intervention.initialize(self)
        self.n_products_used = hpb.Result(name=f'Products administered by {self.label}', npts=sim.res_npts, scale=True)
        self.outcomes = {k: np.array([], dtype=hpd.default_int) for k in ['unsuccessful', 'successful']} # Store outcomes on each timestep

    def check_eligibility(self, sim):
        '''
        Check people's eligibility for treatment
        '''
        females         = sim.people.is_female
        in_age_range    = (sim.people.age >= self.age_range[0]) * (sim.people.age <= self.age_range[1])
        alive           = sim.people.alive
        nocancer        = ~sim.people.cancerous.any(axis=0)
        conditions      = (females * in_age_range * alive * nocancer)
        return conditions

    def get_accept_inds(self, sim):
        '''
        Get indices of people who will acccept treatment; these people are then added to a queue or scheduled for receiving treatment
        '''
        accept_inds     = np.array([], dtype=hpd.default_int)
        is_eligible     = self.check_eligibility(sim) # Apply eligiblity
        if len(self.eligibility(sim)):
            eligible_inds   = hpu.itruei(is_eligible, sc.promotetoarray(self.eligibility(sim)))
            if len(eligible_inds):
                accept_inds     = select_people(eligible_inds, prob=self.prob[0])  # Select people who accept
        return accept_inds

    def get_candidates(self, sim):
        '''
        Get candidates for treatment on this timestep. Implemented by derived classes.
        '''
        raise NotImplementedError

    def apply(self, sim):
        '''
        Perform treatment by getting candidates, checking their eligibility, and then treating them.
        '''
        # Get indices of who will get treated
        treat_candidates = self.get_candidates(sim) # NB, this needs to be implemented by derived classes
        still_eligible = self.check_eligibility(sim)
        treat_inds = hpu.itruei(still_eligible, treat_candidates)

        # Store treatment and dates
        sim.people.cin_treated[treat_inds] = True
        sim.people.cin_treatments[treat_inds] += 1
        sim.people.date_cin_treated[treat_inds] = sim.t

        # Store results
        idx = int(sim.t / sim.resfreq)
        new_treat_inds = hpu.ifalsei(sim.people.cin_treated, treat_inds)  # Figure out people who are getting radiation for the first time
        n_new_cin_treatments = sim.people.scale_flows(treat_inds)  # Scale
        n_new_people = sim.people.scale_flows(new_treat_inds)  # Scale
        sim.results['new_cin_treated'][idx] += n_new_people
        sim.results['new_cin_treatments'][idx] += n_new_cin_treatments

        # Administer treatment and store products used
        self.outcomes = self.product.administer(sim, treat_inds)
        self.n_products_used[idx] += sim.people.scale_flows(treat_inds)

        return treat_inds


class treat_num(BaseTreatment):
    '''
    Treat a fixed number of people each timestep.

    Args:
         max_capacity (int): maximum number who can be treated each timestep
    '''
    def __init__(self, max_capacity=None, **kwargs):
        BaseTreatment.__init__(self, **kwargs)
        self.queue = []
        self.max_capacity = max_capacity
        return

    def add_to_queue(self, sim):
        '''
        Add people who are willing to accept treatment to the queue
        '''
        accept_inds = self.get_accept_inds(sim)
        if len(accept_inds): self.queue += accept_inds.tolist()

    def get_candidates(self, sim):
        '''
        Get the indices of people who are candidates for treatment
        '''
        treat_candidates = np.array([], dtype=hpd.default_int)
        if len(self.queue):
            if self.max_capacity is None or (self.max_capacity>len(self.queue)):
                treat_candidates = self.queue[:]
            else:
                treat_candidates = self.queue[:self.max_capacity]
        return sc.promotetoarray(treat_candidates)

    def apply(self, sim):
        '''
        Apply treatment. On each timestep, this method will add eligible people who are willing to accept treatment to a
        queue, and then will treat as many people in the queue as there is capacity for.
        '''
        self.add_to_queue(sim)
        treat_inds = BaseTreatment.apply(self, sim) # Apply method from BaseTreatment class
        self.queue = [e for e in self.queue if e not in treat_inds] # Recreate the queue, removing people who were treated
        return treat_inds


class treat_delay(BaseTreatment):
    '''
    Treat people after a fixed delay

    Args:
         delay (int): years of delay between becoming eligible for treatment and receiving treatment.
    '''
    def __init__(self, delay=None, **kwargs):
        BaseTreatment.__init__(self, **kwargs)
        self.delay = delay or 0
        self.scheduler = defaultdict(list)

    def add_to_schedule(self, sim):
        '''
        Add people who are willing to accept treatment to the treatment scehduler
        '''
        accept_inds = self.get_accept_inds(sim)
        if len(accept_inds): self.scheduler[sim.t] = accept_inds

    def get_candidates(self, sim):
        '''
        Get the indices of people who are candidates for treatment
        '''
        due_time = sim.t-self.delay/sim['dt']
        treat_candidates = np.array([], dtype=hpd.default_int)
        if len(self.scheduler[due_time]):
            treat_candidates = self.scheduler[due_time]
        return treat_candidates

    def apply(self, sim):
        '''
        Apply treatment. On each timestep, this method will add eligible people who are willing to accept treatment to a
        scheduler, and then will treat anyone scheduled for treatment on this timestep.
        '''
        self.add_to_schedule(sim)
        treat_inds = BaseTreatment.apply(self, sim)
        return treat_inds


class BaseTxVx(BaseTreatment):
    '''
    Base class for therapeutic vaccination
    '''
    def __init__(self, **kwargs):
        BaseTreatment.__init__(self, **kwargs)


    def deliver(self, sim):
        '''
        Deliver the intervention. This applies on a single timestep, whereas apply() methods
        apply on every timestep and can selectively call this method.
        '''
        is_eligible = self.check_eligibility(sim) # Apply general eligiblity

        # Apply extra user-defined eligibility conditions, if given
        if self.eligibility is not None:
            extra_conditions = sc.promotetoarray(self.eligibility(sim))

            # Checking self.eligibility() can return either a boolean array of indices. Convert to indices.
            if (len(extra_conditions) == len(is_eligible)) & (len(extra_conditions) > 0):
                if sc.checktype(extra_conditions[0], 'bool'):
                    extra_conditions = hpu.true(extra_conditions)

            # Combine the extra conditions with general eligibility
            if len(extra_conditions)>0:
                eligible_inds = hpu.itruei(is_eligible, extra_conditions)  # First make sure they're generally eligible
            else:
                eligible_inds = np.array([])

        else:
            eligible_inds = hpu.true(is_eligible)

        # Get anyone eligible and apply acceptance rates
        accept_inds = np.array([])
        if len(eligible_inds): # If so, proceed
            accept_inds = select_people(eligible_inds, prob=self.prob[0]) # Select people who accept
            new_vx_inds = hpu.ifalsei(sim.people.tx_vaccinated, accept_inds) # Figure out people who are getting vaccinated for the first time
            n_new_doses  = sim.people.scale_flows(accept_inds) # Scale
            n_new_people = sim.people.scale_flows(new_vx_inds) # Scale

            if n_new_doses:
                self.outcomes = self.product.administer(sim, accept_inds) # Administer
                sim.people.tx_vaccinated[accept_inds] = True
                sim.people.date_tx_vaccinated[accept_inds] = sim.t
                sim.people.txvx_doses[accept_inds] += 1

                idx = int(sim.t / sim.resfreq)
                sim.results['new_tx_vaccinated'][idx] += n_new_people
                sim.results['new_txvx_doses'][idx] += n_new_doses
                self.n_products_used[idx] += n_new_doses

        return accept_inds


class routine_txvx(BaseTxVx, RoutineDelivery):
    '''
    Routine delivery of therapeutic vaccine - an instance of treat_num combined
     with routine delivery. See base classes for a description of input arguments.

    **Examples**::

        txvx1 = hpv.routine_txvx(product='txvx1', prob=0.9, age_range=[25,26], start_year=2030) # Vaccinate 90% of 25yo women every year starting 2025
        txvx2 = hpv.routine_txvx(product='txvx1', prob=np.linspace(0.2,0.8,5), age_range=[25,26], years=np.arange(2030,2035)) # Scale up vaccination over 5 years starting in 2020
    '''

    def __init__(self, product=None, prob=None, age_range=None, eligibility=None,
                 start_year=None, end_year=None, years=None, annual_prob=None, **kwargs):
        BaseTxVx.__init__(self, product=product, age_range=age_range, eligibility=eligibility, **kwargs)
        RoutineDelivery.__init__(self, prob=prob, start_year=start_year, end_year=end_year, years=years, annual_prob=annual_prob)

    def initialize(self, sim):
        RoutineDelivery.initialize(self, sim) # Initialize this first, as it ensures that prob is interpolated properly
        BaseTxVx.initialize(self, sim) # Initialize this next - this actually calls BaseTreatment.initialize() and ensures products will be counted

    def apply(self, sim):
        accept_inds = np.array([])
        if sim.t in self.timepoints:
            accept_inds = self.deliver(sim)
        return accept_inds


class campaign_txvx(BaseTxVx, CampaignDelivery):
    '''
    Campaign delivery of therapeutic vaccine - an instance of treat_num combined
    with campaign delivery. See base classes for a description of input arguments.
    '''

    def __init__(self, product=None, prob=None, age_range=None, eligibility=None,
                 years=None, interpolate=True, annual_prob=None, **kwargs):
        BaseScreening.__init__(self, product=product, age_range=age_range, eligibility=eligibility, **kwargs)
        CampaignDelivery.__init__(self, prob=prob, years=years, interpolate=interpolate, annual_prob=annual_prob)

    def initialize(self, sim):
        CampaignDelivery.initialize(self, sim) # Initialize this first, as it ensures that prob is interpolated properly
        BaseTxVx.initialize(self, sim) # Initialize this next - this actually calls BaseTreatment.initialize() and ensures products will be counted

    def apply(self, sim):
        accept_inds = np.array([])
        if sim.t in self.timepoints:
            accept_inds = self.deliver(sim)
        return accept_inds


class linked_txvx(BaseTxVx):
    '''
    Deliver therapeutic vaccine. This intervention should be used if TxVx delivery
    is linked to another program that determines eligibility, e.g. a screening program.
    Handling of dates is assumed to be handled by the linked intervention.
    '''
    def __init__(self, **kwargs):
        BaseTxVx.__init__(self, **kwargs)

    def apply(self, sim):
        accept_inds = BaseTxVx.deliver(self, sim)
        return accept_inds


#%% Products
__all__ += ['dx', 'tx', 'vx', 'radiation', 'default_dx', 'default_tx', 'default_vx']

class Product(hpb.FlexPretty):
    ''' Generic product implementation '''
    def administer(self, people, inds):
        ''' Adminster a Product - implemented by derived classes '''
        raise NotImplementedError


class dx(Product):
    '''
    Testing products are used within screening and triage. Their fundamental property is that they classify people
    into exactly one result state. They do not change anything about the People.
    '''
    def __init__(self, df, hierarchy=None):
        self.df = df
        self.states = df.state.unique()
        self.genotypes = df.genotype.unique()
        self.ng = len(self.genotypes)

        if hierarchy is None:
            self.hierarchy = df.result.unique() # Hierarchy is drawn from the order in which the outcomes are specified. The last unique item to be specified is the default
        else:
            self.hierarchy = hierarchy # or ['high', 'medium', 'low', 'not detected'], or other

    @property
    def default_value(self):
        return len(self.hierarchy)-1

    def administer(self, sim, inds, return_format='dict'):
        '''
        Administer a testing product.

        Returns:
             if return_format=='array': an array of length len(inds) with integer entries that map each person to one of the result_states
             if return_format=='dict': a dictionary keyed by result_states with values containing the indices of people classified into this state
        '''

        # Pre-fill with the default value, which is set to be the last value in the hierarchy
        results = np.full_like(inds, fill_value=self.default_value, dtype=hpd.default_int)
        people = sim.people

        for state in self.states:
            for g,genotype in sim['genotype_map'].items():

                theseinds = hpu.true(people[state][g, inds])

                # Filter the dataframe to extract test results for people in this state
                df_filter = (self.df.state == state) # filter by state
                if self.ng>1: df_filter = df_filter & (self.df.genotype == genotype) # also filter by genotype, if this test is by genotype
                thisdf = self.df[df_filter] # apply filter to get the results for this state & genotype
                probs = [thisdf[thisdf.result==result].probability.values[0] for result in self.hierarchy] # Pull out the result probabilities in the order specified by the result hierarchy
                # Sort people into one of the possible result states and then update their overall results (aggregating over genotypes)
                this_gtype_results = hpu.n_multinomial(probs, len(theseinds))
                results[theseinds] = np.minimum(this_gtype_results, results[theseinds])

        if return_format=='dict':
            output = {self.hierarchy[i]:inds[results==i] for i in range(len(self.hierarchy))}
        elif return_format=='array':
            output = results

        return output


class tx(Product):
    '''
    Treatment products include anything used to treat cancer or precancer, as well as therapeutic vaccination.
    They change fundamental properties about People, including their prognoses and infectiousness.
    '''
    def __init__(self, df, clearance=0.8, name=None):
        self.df = df
        self.clearance = clearance
        self.name = df.name.unique()[0]
        self.states = df.state.unique()
        self.genotypes = df.genotype.unique()
        self.ng = len(self.genotypes)

    def get_people_in_state(self, state, g, sim):
        '''
        Find people within a given state/genotype. Returns indices
        '''
        if self.ng==1:  theseinds = sim.people[state].any(axis=0)
        else:           theseinds = hpu.true(sim.people[state][g, :])


    def administer(self, sim, inds, return_format='dict'):
        '''
        Loop over treatment states to determine those who are successfully treated and clear infection
        '''

        tx_successful = [] # Initialize list of successfully treated individuals
        people = sim.people

        for state in self.states: # Loop over states
            for g,genotype in sim['genotype_map'].items(): # Loop over genotypes in the sim

                theseinds = inds[hpu.true(people[state][g, inds])] # Extract people for whom this state is true for this genotype]

                if len(theseinds):

                    df_filter = (self.df.state == state)  # Filter by state
                    if self.ng>1: df_filter = df_filter & (self.df.genotype == genotype)
                    thisdf = self.df[df_filter] # apply filter to get the results for this state & genotype

                    # Determine whether treatment is successful
                    efficacy = thisdf.efficacy.values[0]
                    eff_probs = np.full(len(theseinds), efficacy, dtype=hpd.default_float)  # Assign probabilities of treatment success
                    to_eff_treat = hpu.binomial_arr(eff_probs)  # Determine who will have effective treatment
                    eff_treat_inds = theseinds[to_eff_treat]
                    if len(eff_treat_inds):
                        tx_successful += list(eff_treat_inds)
                        people[state][g, eff_treat_inds] = False  # People who get treated have their CINs removed
                        people['episomal'][g, eff_treat_inds] = False  # People who get treated have their CINs removed
                        people['transformed'][g, eff_treat_inds] = False  # People who get treated have their CINs removed
                        people[f'date_{state}'][g, eff_treat_inds] = np.nan
                        people[f'date_transformed'][g, eff_treat_inds] = np.nan
                        people[f'date_cancerous'][g, eff_treat_inds] = np.nan
                        people[f'sev'][g, eff_treat_inds] = np.nan
                        people['date_clearance'][g, eff_treat_inds] = people.t + 1
                        # Determine whether women also clear infection
                        # clearance_probs = np.full(len(eff_treat_inds), self.clearance, dtype=hpd.default_float)
                        # to_clear = hpu.binomial_arr(clearance_probs)  # Determine who will have effective treatment
                        # clear_inds = eff_treat_inds[to_clear]
                        # if len(clear_inds):
                        #     # If so, set date of clearance of infection on next timestep
                        #
                        #     people.dur_infection[g, clear_inds] = (people.t - people.date_infectious[g, clear_inds]) * people.pars['dt']

        tx_successful = np.array(list(set(tx_successful)))
        tx_unsuccessful = np.setdiff1d(inds, tx_successful)
        if return_format=='dict':
            output = {'successful':tx_successful, 'unsuccessful': tx_unsuccessful}
        elif return_format=='array':
            output = tx_successful

        return output


class vx(Product):
    ''' Vaccine product '''
    def __init__(self, genotype_pars=None, imm_init=None, imm_boost=None):
        self.genotype_pars = genotype_pars
        self.imm_init = imm_init
        self.imm_boost = imm_boost
        self.imm_source = None # Set during immunity initialization. Warning, fragile!!!
        if (imm_init is None and imm_boost is None) or (imm_init is not None and imm_boost is not None):
            errormsg = 'Must provide either an initial immune effect (for first doses) or an immune boosting effect (for subsequent doses), not both/neither.'
            raise ValueError(errormsg)


    def administer(self, people, inds):
        ''' Apply the vaccine to the requested people indices. '''
        inds = inds[people.alive[inds]]  # Skip anyone that is dead
        if self.imm_init is not None:
            people.peak_imm[self.imm_source, inds] = hpu.sample(**self.imm_init, size=len(inds))
        elif self.imm_boost is not None:
            people.peak_imm[self.imm_source, inds] *= self.imm_boost
        people.t_imm_event[self.imm_source, inds] = people.t



class radiation(Product):
    # Cancer treatment product
    def __init__(self, dur=None):
        self.dur = dur or dict(dist='normal', par1=18.0, par2=2.) # whatever the default duration should be

    def administer(self, sim, inds):
        people = sim.people
        new_dur_cancer = hpu.sample(**self.dur, size=len(inds))
        people.date_dead_cancer[inds] += np.ceil(new_dur_cancer / people.pars['dt'])

        # Store treatment and dates
        sim.people.cancer_treated[inds] = True
        sim.people.cancer_treatments[inds] += 1
        sim.people.date_cancer_treated[inds] = sim.t

        # Store results
        idx = int(sim.t / sim.resfreq)
        new_cctreat_inds = hpu.ifalsei(sim.people.cancer_treated, inds)  # Figure out people who are getting radiation for the first time
        n_new_radiaitons = sim.people.scale_flows(inds)  # Scale
        n_new_people = sim.people.scale_flows(new_cctreat_inds)  # Scale
        sim.results['new_cancer_treated'][idx] += n_new_people
        sim.results['new_cancer_treatments'][idx] += n_new_radiaitons

        return inds


#%% Create default products

def default_dx(prod_name=None):
    '''
    Create default diagnostic products
    '''
    dfdx = pd.read_csv(datafiles.dx) # Read in dataframe with parameters
    dxprods = dict(
        # Default primary screening diagnostics
        via             = dx(dfdx[dfdx.name == 'via'],              hierarchy=['positive', 'inadequate', 'negative']),
        lbc             = dx(dfdx[dfdx.name == 'lbc'],              hierarchy=['abnormal', 'ascus', 'inadequate', 'normal']),
        pap             = dx(dfdx[dfdx.name == 'pap'],              hierarchy=['abnormal', 'ascus', 'inadequate', 'normal']),
        colposcopy      = dx(dfdx[dfdx.name == 'colposcopy'],       hierarchy=['cancer', 'hsil', 'lsil', 'ascus', 'normal']),
        hpv             = dx(dfdx[dfdx.name == 'hpv'],              hierarchy=['positive', 'inadequate', 'negative']),
        hpv1618         = dx(dfdx[dfdx.name == 'hpv1618'],          hierarchy=['positive', 'inadequate', 'negative']),
        hpv_type        = dx(dfdx[dfdx.name == 'hpv_type'],         hierarchy=['positive_1618', 'positive_ohr', 'inadequate', 'negative']),
        # Diagnostics used to determine of subsequent care pathways
        txvx_assigner   = dx(dfdx[dfdx.name == 'txvx_assigner'],    hierarchy=['triage', 'txvx', 'none']),
        tx_assigner     = dx(dfdx[dfdx.name == 'tx_assigner'],      hierarchy=['radiation', 'excision', 'ablation', 'none']),
    )
    if prod_name is not None:   return dxprods[prod_name]
    else:                       return dxprods


def default_tx(prod_name=None):
    '''
    Create default treatment products
    '''
    dftx = pd.read_csv(datafiles.tx) # Read in dataframe with parameters
    txprods = dict()
    for name in dftx.name.unique():
        txprods[name] = tx(dftx[dftx.name==name])
    if prod_name is not None:   return txprods[prod_name]
    else:                       return txprods


def default_vx(prod_name=None):
    '''
    Create default vaccine products
    '''
    dfvx = pd.read_csv(datafiles.vx) # Read in dataframe with parameters
    vxprods = dict()
    for name in dfvx.name.unique():
        vxprods[name]       = vx(genotype_pars=dfvx[dfvx.name==name], imm_init=dict(dist='beta', par1=30, par2=2))
        vxprods[name+'2']   = vx(genotype_pars=dfvx[dfvx.name==name], imm_boost=1.2) # 2nd dose
        vxprods[name+'3']   = vx(genotype_pars=dfvx[dfvx.name==name], imm_boost=1.1) # 3rd dose
    if prod_name is not None:   return vxprods[prod_name]
    else:                       return vxprods



'''
Define core Sim classes
'''

# Imports
import numpy as np
import pandas as pd
import sciris as sc
from . import base as hpb
from . import misc as hpm
from . import defaults as hpd
from . import utils as hpu
from . import population as hppop
from . import parameters as hppar
from . import analysis as hpa
from . import plotting as hpplt
from . import immunity as hpimm
from . import interventions as hpi
from .settings import options as hpo


# Define the model
class Sim(hpb.BaseSim):

    def __init__(self, pars=None, datafile=None, label=None,
                 popfile=None, people=None, version=None, hiv_datafile=None, art_datafile=None, **kwargs):

        # Set attributes
        self.label         = label    # The label/name of the simulation
        self.created       = None     # The datetime the sim was created
        self.datafile      = datafile # The name of the data file
        self.art_datafile  = art_datafile # The name of the ART data file
        self.hiv_datafile  = hiv_datafile # The name of the HIV data file
        self.popfile       = popfile  # The population file
        self.data          = None     # The data
        self.popdict       = people   # The population dictionary
        self.people        = None     # Initialize these here so methods that check their length can see they're empty
        self.t             = None     # The current time in the simulation (during execution); outside of sim.step(), its value corresponds to next timestep to be computed
        self.results       = {}       # For storing results
        self.summary       = None     # For storing a summary of the results
        self.initialized   = False    # Whether or not initialization is complete
        self.complete      = False    # Whether a simulation has completed running
        self.results_ready = False    # Whether or not results are ready
        self._default_ver  = version  # Default version of parameters used
        self._orig_pars    = None     # Store original parameters to optionally restore at the end of the simulation

        # Make default parameters (using values from parameters.py)
        default_pars = hppar.make_pars(version=version) # Start with default pars
        super().__init__(default_pars) # Initialize and set the parameters as attributes

        # Load data, including datafile that are used to create additional optional parameters
        location = pars.get('location') if pars else None
        self.load_data(datafile) # Load the data, if provided
        self.load_hiv_data(location=location, hiv_datafile=hiv_datafile, art_datafile=art_datafile) # Load any data that's used to create additional parameters (thus far, HIV and ART)

        # Update parameters
        self.update_pars(pars, **kwargs)   # Update the parameters

        return


    def load_data(self, datafile=None, **kwargs):
        ''' Load the data to calibrate against, if provided '''
        if datafile is not None: # If a data file is provided, load it
            self.data = hpm.load_data(datafile=datafile, check_date=True, **kwargs)
        return


    def load_hiv_data(self, location=None, hiv_datafile=None, art_datafile=None, **kwargs):
        ''' Load any data files that are used to create additional parameters, if provided '''
        self.hiv_data = sc.objdict()
        self.hiv_data.infection_rates, self.hiv_data.art_adherence = hppar.get_hiv_data(location=location, hiv_datafile=hiv_datafile, art_datafile=art_datafile)
        return


    def initialize(self, reset=False, init_states=True, **kwargs):
        '''
        Perform all initializations on the sim.
        '''
        self.t = 0  # The current time index
        self.validate_pars() # Ensure parameters have valid values
        hpu.set_seed(self['rand_seed']) # Reset the random seed before the population is created
        self.init_genotypes() # Initialize the genotypes
        self.init_results() # After initializing the genotypes and people, create the results structure
        self.init_interventions()  # Initialize the interventions BEFORE the people, because then vaccination interventions get counted in immunity structures
        self.init_immunity() # initialize information about immunity
        self.init_people(reset=reset, init_states=init_states, **kwargs) # Create all the people (the heaviest step)
        self.init_analyzers()  # ...and the analyzers...
        hpu.set_seed(self['rand_seed']+1)  # Reset the random seed to the default run seed, so that if the simulation is run with reset_seed=False right after initialization, it will still produce the same output
        self.initialized   = True
        self.complete      = False
        self.results_ready = False

        return self


    def layer_keys(self):
        '''
        Attempt to retrieve the current layer keys.
        '''
        try:
            keys = list(self['acts'].keys()) # Get keys from acts
        except: # pragma: no cover
            keys = []
        return keys


    def reset_layer_pars(self, layer_keys=None, force=False):
        '''
        Reset the parameters to match the population.

        Args:
            layer_keys (list): override the default layer keys (use stored keys by default)
            force (bool): reset the parameters even if they already exist
        '''
        if layer_keys is None:
            if self.people is not None: # If people exist
                layer_keys = self.people.contacts.keys()
            elif self.popdict is not None:
                layer_keys = self.popdict['layer_keys']
        hppar.reset_layer_pars(self.pars, layer_keys=layer_keys, force=force)
        return


    def validate_layer_pars(self):
        '''
        Handle layer parameters, since they need to be validated after the population
        creation, rather than before.
        '''

        # First, try to figure out what the layer keys should be and perform basic type checking
        layer_keys = self.layer_keys()
        layer_pars = hppar.layer_pars # The names of the parameters that are specified by layer
        for lp in layer_pars:
            val = self[lp]
            if sc.isnumber(val): # It's a scalar instead of a dict, assume it's all contacts
                self[lp] = {k:val for k in layer_keys}

        # Handle key mismatches
        for lp in layer_pars:
            lp_keys = set(self.pars[lp].keys())
            if lp != 'layer_probs':
                if not lp_keys == set(layer_keys):
                    errormsg = 'At least one layer parameter is inconsistent with the layer keys; all parameters must have the same keys:'
                    errormsg += f'\nsim.layer_keys() = {layer_keys}'
                    for lp2 in layer_pars: # Fail on first error, but re-loop to list all of them
                        errormsg += f'\n{lp2} = ' + ', '.join(self.pars[lp2].keys())
                    raise sc.KeyNotFoundError(errormsg)

            # TODO: add validation here for layer_probs

        # Handle mismatches with the population
        if self.people is not None:
            pop_keys = set(self.people.contacts.keys())
            if pop_keys != set(layer_keys): # pragma: no cover
                if not len(pop_keys):
                    errormsg = f'Your population does not have any layer keys, but your simulation does {layer_keys}. If you called hpv.People() directly, you probably need hpv.make_people() instead.'
                    raise sc.KeyNotFoundError(errormsg)
                else:
                    errormsg = f'Please update your parameter keys {layer_keys} to match population keys {pop_keys}. You may find sim.reset_layer_pars() helpful.'
                    raise sc.KeyNotFoundError(errormsg)

        return


    def validate_pars(self, validate_layers=True):
        '''
        Some parameters can take multiple types; this makes them consistent.

        Args:
            validate_layers (bool): whether to validate layer parameters as well via validate_layer_pars() -- usually yes, except during initialization
        '''

        # Handle types
        for key in ['n_agents']:
            try:
                self[key] = int(self[key])
            except Exception as E:
                errormsg = f'Could not convert {key}={self[key]} of {type(self[key])} to integer'
                raise ValueError(errormsg) from E

        # Handle start
        if self['start'] in [None, 0]: # Use default start
            self['start'] = 2015

        # Handle end and n_years
        if self['end']:
            self['n_years'] = int(self['end'] - self['start'])
            if self['n_years'] <= 0:
                errormsg = f"Number of years must be >0, but you supplied start={str(self['start'])} and end={str(self['end'])}, which gives n_years={self['n_years']}"
                raise ValueError(errormsg)
        else:
            if self['n_years']:
                self['end'] = self['start'] + self['n_years']
            else:
                errormsg = 'You must supply one of n_years and end."'
                raise ValueError(errormsg)

        # Construct other things that keep track of time
        self.years      = sc.inclusiverange(self['start'],self['end'])
        self.yearvec    = sc.inclusiverange(start=self['start'], stop=self['end']+1-self['dt'], step=self['dt']) # Includes all the timepoints in the last year
        self.npts       = len(self.yearvec)
        self.tvec       = np.arange(self.npts)

        # Handle population network data
        network_choices = ['random', 'default']
        choice = self['network']
        if choice and choice not in network_choices: # pragma: no cover
            choicestr = ', '.join(network_choices)
            errormsg = f'Population type "{choice}" not available; choices are: {choicestr}'
            raise ValueError(errormsg)

        # Handle analyzers and interventions
        for key in ['interventions', 'analyzers']: # Ensure all of them are lists
            self[key] = sc.dcp(sc.tolist(self[key], keepnone=False)) # All of these have initialize functions that run into issues if they're reused
        for i,interv in enumerate(self['interventions']):
            if isinstance(interv, dict): # It's a dictionary representation of an intervention
                self['interventions'][i] = hpi.InterventionDict(**interv)

        # Optionally handle layer parameters
        if validate_layers:
            self.validate_layer_pars()

        # Handle verbose
        if self['verbose'] == 'brief':
            self['verbose'] = -1
        if not sc.isnumber(self['verbose']): # pragma: no cover
            errormsg = f'Verbose argument should be either "brief", -1, or a float, not {type(self["verbose"])} "{self["verbose"]}"'
            raise ValueError(errormsg)

        # Handle HIV
        if self['model_hiv']:
            if self.hiv_data['infection_rates'] is None or self.hiv_data['art_adherence'] is None:
                errormsg = 'Data on HIV infection rates and ART adherence must be provided if model_hiv is True.'
                raise ValueError(errormsg)

        return


    def validate_init_conditions(self, init_hpv_prev):
        '''
        Initial prevalence values can be supplied with different amounts of detail.
        Here we flesh out any missing details so that the initial prev values are
        by age and genotype. We also check the prevalence values are ok.
        '''

        def validate_arrays(vals, n_age_brackets=None):
            ''' Little helper function to check prevalence values '''
            if n_age_brackets is not None:
                if len(vals) != n_age_brackets:
                    errormsg = f'The initial prevalence values must either be the same length as the age brackets: {len(vals)} vs {n_age_brackets}.'
                    raise ValueError(errormsg)
            else:
                if len(vals) != 1:
                    errormsg = f'No age brackets were supplied, but more than one prevalence value was supplied ({len(vals)}). An array of prevalence values can only be supplied along with an array of corresponding age brackets.'
                    raise ValueError(errormsg)
            if vals.any() < 0 or vals.any() > 1:
                errormsg = f'The initial prevalence values must either between 0 and 1, not {vals}.'
                raise ValueError(errormsg)

            return

        # If values have been provided, validate them
        sex_keys = {'m', 'f'}
        tot_keys = ['all', 'total', 'tot', 'average', 'avg']
        n_age_brackets = None

        if init_hpv_prev is not None:
            if sc.checktype(init_hpv_prev, dict):
                # Get age brackets if supplied
                if 'age_brackets' in init_hpv_prev.keys():
                    age_brackets = init_hpv_prev.pop('age_brackets')
                    n_age_brackets = len(age_brackets)
                else:
                    age_brackets = np.array([150])

                # Handle the rest of the keys
                var_keys = list(init_hpv_prev.keys())
                if (len(var_keys)==1 and var_keys[0] not in tot_keys) or (len(var_keys)>1 and set(var_keys) != sex_keys):
                    errormsg = f'Could not understand the initial prevalence provided: {init_hpv_prev}. If supplying a dictionary, please use "m" and "f" keys or "tot". '
                    raise ValueError(errormsg)
                if len(var_keys) == 1:
                    k = var_keys[0]
                    init_hpv_prev = {sk: sc.promotetoarray(init_hpv_prev[k]) for sk in sex_keys}

                # Now set the values
                for k, vals in init_hpv_prev.items():
                    init_hpv_prev[k] = sc.promotetoarray(vals)

            elif sc.checktype(init_hpv_prev, 'arraylike') or sc.isnumber(init_hpv_prev):
                # If it's an array, assume these values apply to males and females
                init_hpv_prev = {sk: sc.promotetoarray(init_hpv_prev) for sk in sex_keys}
                age_brackets = np.array([150])

            else:
                errormsg = f'Initial prevalence values of type {type(init_hpv_prev)} not recognized, must be a dict, an array, or a float.'
                raise ValueError(errormsg)

            # Now validate the arrays
            for sk, vals in init_hpv_prev.items():
                validate_arrays(vals, n_age_brackets)

        # If values haven't been supplied, assume zero
        else:
            init_hpv_prev = {'f': np.array([0]), 'm': np.array([0])}
            age_brackets = np.array([150])

        return init_hpv_prev, age_brackets


    def init_genotypes(self):
        ''' Initialize the genotype parameters '''
        if self._orig_pars and 'genotypes' in self._orig_pars:
            self['genotypes'] = self._orig_pars.pop('genotypes')  # Restore

        default_gpars   = hppar.get_genotype_pars()
        user_gpars      = sc.dcp(self['genotype_pars'])
        self['genotype_pars'] = sc.objdict()

        # Handle special input cases
        if self['genotypes'] == 'all':
            self['genotypes'] = default_gpars.keys()
        if not len(self['genotypes']):
            print('No genotypes provided, will simulate 16, 18, and other HR types by default')
            self['genotypes'] = [16,18,'hrhpv']

        # Loop over genotypes
        for i, g in enumerate(self['genotypes']):

            # Standardize format of genotype inputs
            if sc.isnumber(g): g = f'hpv{g}' # Convert e.g. 16 to hpv16
            if sc.checktype(g,str):
                if not g in default_gpars.keys():
                    errormsg = f'Genotype {i} ({g}) is not one of the inbuilt options.'
                    raise ValueError(errormsg)
            else:
                errormsg = f'Format {type(g)} is not understood.'
                raise ValueError(errormsg)

            # Add to genotype_par dict
            self['genotype_pars'][g] = default_gpars[g]
            self['genotype_map'][i] = g

        # Loop over user-supplied genotype parameters that can overwrite values
        if len(user_gpars):
            for g,gpars in user_gpars.items():

                # Standardize format of genotype inputs
                if sc.isnumber(g): g = f'hpv{g}'  # Convert e.g. 16 to hpv16
                if sc.checktype(g, str):
                    if not g in self['genotype_pars'].keys():
                        errormsg = f'Parameters provided for genotype {g}, but it is not in the sim.'
                        raise ValueError(errormsg)
                    else:
                        for gparname,gparval in gpars.items():
                            if gparname in self['genotype_pars'][g].keys():
                                printmsg = f"Resetting parameter '{gparname}' from {self['genotype_pars'][g][gparname]} to {gparval} for genotype {g}"
                                sc.printv(printmsg, 1, self['verbose'])
                                self['genotype_pars'][g][gparname] = gparval
                            else:
                                errormsg = f"Parameter {gparname} does not exist for genotype {g}"
                                raise ValueError(errormsg)

        len_pars = len(self['genotype_pars'])
        self['n_genotypes'] = len_pars  # Each genotype has an entry in genotype_pars

        # Set the number of immunity sources
        self['n_imm_sources'] = len(self['genotypes'])

        return

    def init_immunity(self, create=True):
        ''' Initialize immunity matrices '''
        hpimm.init_immunity(self, create=create)
        return


    def init_results(self, frequency='annual', add_data=True):
        '''
        Create the main results structure.
        The prefix "n" is used for stock variables, i.e. counting the total number in any given state (sus/inf/etc) on any particular timestep

        Arguments:
            sim         (hpv.Sim)       : a sim
            frequency   (str or float)  : the frequency with which to save results: accepts 'annual', 'dt', or a float which is interpreted as a fraction of a year, e.g. 0.2 will save results every 0.2 years
            add_data    (bool)          : whether or not to add data to the result structures
        '''

        # Handle frequency
        if type(frequency) == str:
            if frequency == 'annual':
                resfreq = int(1 / self['dt'])
            elif frequency == 'dt':
                resfreq = 1
            else:
                errormsg = f'Result frequency not understood: must be "annual", "dt" or a float, but you provided {frequency}.'
                raise ValueError(errormsg)
        elif type(frequency) == float:
            if frequency < self['dt']:
                errormsg = f'You requested results with frequency {frequency}, but this is smaller than the simulation timestep {self["dt"]}.'
                raise ValueError(errormsg)
            else:
                resfreq = int(frequency / self['dt'])
        self.resfreq = resfreq
        if not self.resfreq > 0:
            errormsg = f'The results frequence should be a positive integer, not {self.resfreq}: dt may be too large'
            raise ValueError(errormsg)

        # Construct the tvec that will be used with the results
        points_to_use = np.arange(0, self.npts, self.resfreq)
        self.res_yearvec = self.yearvec[points_to_use]
        self.res_npts = len(self.res_yearvec)
        self.res_tvec = np.arange(self.res_npts)

        # Function to create results
        def init_res(*args, **kwargs):
            ''' Initialize a single result object '''
            output = hpb.Result(*args, **kwargs, npts=self.res_npts)
            return output

        # Initialize storage
        results = sc.objdict()

        ng = self['n_genotypes'] # Number of genotypes
        na = len(self['age_bins']) - 1 # Number of age bins

        # Create flows
        for flow in hpd.flows:
            results[flow.name]                  = init_res(flow.label, color=flow.color)
            results[flow.name+'_by_genotype']   = init_res(flow.label+' by genotype', n_rows=ng)
            results[flow.name+'_by_age']        = init_res(flow.label+' by age', n_rows=na, color=flow.color)

        # Create stocks
        for stock in hpd.PeopleMeta.stock_states:
            results[f'n_{stock.name}']              = init_res(stock.label, color=stock.color)
            results[f'n_{stock.name}_by_genotype']  = init_res(stock.label+' by genotype', n_rows=ng)

        # Only by-age stock result we will need is number infectious, susceptible, and with cin, for HPV and CIN prevalence/incidence calculations
        results[f'n_infectious_by_age']             = init_res('Number infectious by age', n_rows=na, color=stock.color)
        results[f'n_susceptible_by_age']            = init_res('Number susceptible by age', n_rows=na, color=stock.color)
        results[f'n_transformed_by_age']            = init_res('Number transformed by age', n_rows=na, color=stock.color)
        results[f'n_precin_by_age']                 = init_res('Number Pre-CIN by age', n_rows=na, color=stock.color)
        results[f'n_cin1_by_age']                   = init_res('Number CIN1 by age', n_rows=na, color=stock.color)
        results[f'n_cin2_by_age']                   = init_res('Number CIN2 by age', n_rows=na, color=stock.color)
        results[f'n_cin3_by_age']                   = init_res('Number CIN3 by age', n_rows=na, color=stock.color)

        # Create incidence and prevalence results
        for var,name,color in zip(hpd.inci_keys, hpd.inci_names, hpd.inci_colors):
            results[f'{var}_incidence']             = init_res(name+' incidence', color=color)
            results[f'{var}_incidence_by_genotype'] = init_res(name+' incidence by genotype', n_rows=ng)
            results[f'{var}_incidence_by_age']      = init_res(name+' incidence by age', n_rows=na, color=color)

        # Create demographic flows
        for var, name, color in zip(hpd.dem_keys, hpd.dem_names, hpd.dem_colors):
            results[var] = init_res(name, color=color)

        # Create results by sex
        for var, name, color in zip(hpd.by_sex_keys, hpd.by_sex_colors, hpd.by_sex_colors):
            results[var] = init_res(name, color=color, n_rows=2)

        # Create ASR results using standard populations
        results['asr_cancer_incidence'] = init_res('Age-adjusted cervical cancer incidence', scale=False)
        results['asr_cancer_mortality'] = init_res('Age-adjusted cervical cancer mortality', scale=False)

        stock_colors = [i for i in set(hpd.stock_colors) if i is not None]

        # Type distributions by cytology
        for var, name in zip(hpd.type_dist_keys, hpd.type_dist_names):
            results[var+'_genotype_dist'] = init_res(name, n_rows=ng, color=stock_colors[0])

        # Vaccination results
        results['new_vaccinated'] = init_res('Newly vaccinated by genotype', n_rows=ng)
        results['new_total_vaccinated'] = init_res('Newly vaccinated')
        results['cum_vaccinated'] = init_res('Cumulative number vaccinated by genotype', n_rows=ng)
        results['cum_total_vaccinated'] = init_res('Cumulative number vaccinated')
        results['new_doses'] = init_res('New doses')
        results['cum_doses'] = init_res('Cumulative doses')

        # Therapeutic vaccine results
        results['new_txvx_doses'] = init_res('New therapeutic vaccine doses')
        results['new_tx_vaccinated'] = init_res('Newly received therapeutic vaccine')
        results['cum_txvx_doses'] = init_res('Cumulative therapeutic vaccine doses')
        results['cum_tx_vaccinated'] = init_res('Total received therapeutic vaccine')

        # Screen & treat results
        results['new_screens'] = init_res('New screens')
        results['new_screened'] = init_res('Newly screened')
        results['new_cin_treatments'] = init_res('New CIN treatments')
        results['new_cin_treated'] = init_res('Newly treated for CINs')
        results['new_cancer_treatments'] = init_res('New cancer treatments')
        results['new_cancer_treated'] = init_res('Newly treated for cancer')
        results['cum_screens'] = init_res('Cumulative screens')
        results['cum_screened'] = init_res('Cumulative number screened')
        results['cum_cin_treatments'] = init_res('Cumulative CIN treatments')
        results['cum_cin_treated'] = init_res('Cumulative number treated for CINs')
        results['cum_cancer_treatments'] = init_res('Cumulative cancer treatments')
        results['cum_cancer_treated'] = init_res('Cumulative number treated for cancer')

        # Additional cancer results
        results['detected_cancer_incidence'] = init_res('Detected cancer incidence', color='#fcba03')
        results['cancer_mortality'] = init_res('Cancer mortality')

        # Other results
        results['n_alive'] = init_res('Number alive')
        results['n_alive_by_sex'] = init_res('Number alive by sex', n_rows=2)
        results['n_alive_by_age'] = init_res('Number alive by age', n_rows=na)
        results['n_females_alive_by_age'] = init_res('Number females alive by age', n_rows=na)
        results['cdr'] = init_res('Crude death rate', scale=False)
        results['cbr'] = init_res('Crude birth rate', scale=False, color='#fcba03')
        results['hiv_incidence'] = init_res('HIV incidence')
        results['hiv_prevalence'] = init_res('HIV prevalence')
        results['hpv_prevalence'] = init_res('HPV prevalence', color=stock_colors[0])
        results['hpv_prevalence_by_genotype'] = init_res('HPV prevalence', n_rows=ng, color=stock_colors[0])
        results['hpv_prevalence_by_age'] = init_res('HPV prevalence by age', n_rows=na, color=stock_colors[0])
        results['precin_prevalence'] = init_res('Pre-CIN prevalence', color=stock_colors[0])
        results['precin_prevalence_by_genotype'] = init_res('Pre-CIN prevalence by genotype', n_rows=ng, color=stock_colors[0])
        results['precin_prevalence_by_age'] = init_res('Pre-CIN prevalence by age', n_rows=na, color=stock_colors[0])
        results['cin1_prevalence'] = init_res('CIN1 prevalence', color=stock_colors[1])
        results['cin1_prevalence_by_genotype'] = init_res('CIN1 prevalence by genotype', n_rows=ng, color=stock_colors[1])
        results['cin1_prevalence_by_age'] = init_res('CIN1 prevalence by age', n_rows=na, color=stock_colors[1])
        results['cin2_prevalence'] = init_res('CIN2 prevalence', color=stock_colors[2])
        results['cin2_prevalence_by_genotype'] = init_res('CIN2 prevalence by genotype', n_rows=ng, color=stock_colors[2])
        results['cin2_prevalence_by_age'] = init_res('CIN2 prevalence by age', n_rows=na, color=stock_colors[2])
        results['cin3_prevalence'] = init_res('CIN3 prevalence', color=stock_colors[3])
        results['cin3_prevalence_by_genotype'] = init_res('CIN3 prevalence', n_rows=ng, color=stock_colors[3])
        results['cin3_prevalence_by_age'] = init_res('CIN3 prevalence by age', n_rows=na, color=stock_colors[3])

        # Time vector
        results['year'] = self.res_yearvec
        results['t'] = self.res_tvec

        # Final items
        self.results = results
        self.results_ready = False

        return


    def init_people(self, popdict=None, init_states=False, reset=False, verbose=None, **kwargs):
        '''
        Create the people and the network.

        Use ``init_states=False`` for creating a fresh People object for use
        in future simulations

        Args:
            popdict         (any):  pre-generated people of various formats.
            init_states     (bool): whether to initialize states (default false when called directly)
            reset           (bool): whether to regenerate the people even if they already exist
            verbose         (int):  detail to print
            kwargs          (dict): passed to hpv.make_people()
        '''

        # Handle inputs
        if verbose is None:
            verbose = self['verbose']
        if popdict is not None:
            self.popdict = popdict
        if verbose > 0:
            resetstr= ''
            if self.people:
                resetstr = ' (resetting people)' if reset else ' (warning: not resetting sim.people)'
            print(f'Initializing sim{resetstr} with {self["n_agents"]:0n} agents')
        if self.popfile and self.popdict is None: # If there's a popdict, we initialize it
            self.load_population(init_people=False)

        # Actually make the people
        self.people, total_pop = hppop.make_people(self, reset=reset, verbose=verbose, microstructure=self['network'], **kwargs)
        
        # Figure out the scale factors
        if self['total_pop'] is not None and total_pop is not None: # If no pop_scale has been provided, try to get it from the location
            errormsg = 'You can either define total_pop explicitly or via the location, but not both'
            raise ValueError(errormsg)
        elif total_pop is None and self['total_pop'] is not None:
            total_pop = self['total_pop']
            
        if self['pop_scale'] is None:
            if total_pop is None:
                self['pop_scale'] = 1.0
            else:
                self['pop_scale'] = total_pop/self['n_agents']
        self['ms_agent_ratio'] = int(self['ms_agent_ratio'])
        
        # Finish initialization
        self.people.initialize(sim_pars=self.pars, hiv_pars=self.hiv_data) # Fully initialize the people
        self.reset_layer_pars(force=False) # Ensure that layer keys match the loaded population
        if init_states:
            init_hpv_prev = sc.dcp(self['init_hpv_prev'])
            init_hpv_prev, age_brackets = self.validate_init_conditions(init_hpv_prev)
            self.init_states(age_brackets=age_brackets, init_hpv_prev=init_hpv_prev)

        return self


    def init_interventions(self):
        ''' Initialize and validate the interventions '''

        # Initialization
        self.interventions = sc.autolist()

        # Translate the intervention specs into actual interventions
        for i,intervention in enumerate(self['interventions']):
            if isinstance(intervention, type) and issubclass(intervention, hpi.Intervention):
                intervention = intervention() # Convert from a class to an instance of a class
            if isinstance(intervention, hpi.Intervention):
                intervention.initialize(self)
                self.interventions += intervention
            elif callable(intervention):
                self.interventions += intervention
            else:
                errormsg = f'Intervention {intervention} does not seem to be a valid intervention: must be a function or hpv.Intervention subclass'
                raise TypeError(errormsg)

        return


    def init_analyzers(self):
        ''' Initialize the analyzers '''

        self.analyzers = sc.autolist()

        def convert_analyzer(analyzer):
            ''' Helper function to turn strings into analyzers '''
            choices = hpa.analyzer_map.keys()
            if not analyzer in choices:
                errormsg = f'Analyzer {analyzer} not understood: choices are {choices}.'
                raise ValueError(errormsg)
            else:
                analyzer = hpa.analyzer_map[analyzer]
            return analyzer

        # Interpret analyzers
        for ai,analyzer in enumerate(self['analyzers']):
            if isinstance(analyzer, str):
                analyzer_list = sc.tolist(convert_analyzer(analyzer)) # If not a list, turn it into one - for consistency of processing
                for az in analyzer_list:
                    if isinstance(az, str): az = convert_analyzer(az) # It might still be a string
                    self.analyzers += az() # Unpack list
            else:
                if isinstance(analyzer, type) and issubclass(analyzer, hpa.Analyzer):
                    analyzer = analyzer() # Convert from a class to an instance of a class
                if not (isinstance(analyzer, hpa.Analyzer) or callable(analyzer)):
                    errormsg = f'Analyzer {analyzer} does not seem to be a valid analyzer: must be a function or hpv.Analyzer subclass'
                    raise TypeError(errormsg)
                self.analyzers += analyzer # Add it in

        for analyzer in self.analyzers:
            if isinstance(analyzer, hpa.Analyzer):
                analyzer.initialize(self)
                
        return


    def finalize_analyzers(self):
        for analyzer in self.analyzers:
            if isinstance(analyzer, hpa.Analyzer):
                analyzer.finalize(self)


    def init_states(self, age_brackets=None, init_hpv_prev=None, init_cin_prev=None, init_cancer_prev=None):
        '''
        Initialize prior immunity and seed infections
        '''

        # Shorten key variables
        ng = self['n_genotypes']

        # Assign people to age buckets
        age_inds = np.digitize(self.people.age, age_brackets)

        # Assign probabilities of having HPV to each age/sex group
        hpv_probs = np.full(len(self.people), np.nan, dtype=hpd.default_float)
        hpv_probs[self.people.f_inds] = init_hpv_prev['f'][age_inds[self.people.f_inds]]*self.pars['rel_init_prev']
        hpv_probs[self.people.m_inds] = init_hpv_prev['m'][age_inds[self.people.m_inds]]*self.pars['rel_init_prev']
        hpv_probs[~self.people.is_active] = 0 # Blank out people who are not yet sexually active

        # Get indices of people who have HPV
        hpv_inds = hpu.true(hpu.binomial_arr(hpv_probs))

        # Determine which genotype people are infected with
        if self['init_hpv_dist'] is None: # No type distribution provided, assume even split
            genotypes = np.random.randint(0, ng, len(hpv_inds))
        else:
            # Error checking
            if not sc.checktype(self['init_hpv_dist'], dict):
                errormsg = f'Please provide initial HPV type distribution as a dictionary keyed by genotype, not {self["init_hpv_dist"]}'
                raise ValueError(errormsg)
            if set(self['init_hpv_dist'].keys())!=set(self['genotype_map'].values()):
                errormsg = f'The HPV types provided in the initial HPV type distribution are not the same as the HPV types being simulated: {self["init_hpv_dist"].keys()} vs {self["genotype_map"].values()}.'
                raise ValueError(errormsg)

            type_dist = np.array(list(self['init_hpv_dist'].values()))
            genotypes = hpu.choose_w(type_dist, len(hpv_inds), unique=False)

        for g in range(ng):
            self.people.infect(inds=hpv_inds[genotypes==g], g=g, layer='seed_infection')

        return


    def step(self):
        ''' Step through time and update values '''

        # Set the time and if we have reached the end of the simulation, then do nothing
        if self.complete:
            raise AlreadyRunError('Simulation already complete (call sim.initialize() to re-run)')

        # Shorten key variables
        dt = self['dt'] # Timestep
        t = self.t
        ng = self['n_genotypes']
        na = len(self.pars['age_bins']) - 1 # Number of age bins
        condoms = self['condoms']
        eff_condoms = self['eff_condoms']
        beta = self['beta']
        gen_pars = self['genotype_pars']
        imm_kin_pars = self['imm_kin']
        mixing = self['mixing']
        layer_probs = self['layer_probs']
        cross_layer = self['cross_layer']
        acts = self['acts']
        dur_pship = self['dur_pship']
        age_act_pars = self['age_act_pars']
        trans = np.array([self['transf2m'],self['transm2f']]) # F2M first since that's the order things are done later

        # Update demographics, states, and partnerships
        self.people.update_states_pre(t=t, year=self.yearvec[t]) # This also ages people, applies deaths, and generates new births
        people = self.people # Shorten
        people.dissolve_partnerships(t=t) # Dissolve partnerships
        tind = self.yearvec[t] - self['start']
        people.create_partnerships(tind, mixing, layer_probs, cross_layer, dur_pship, acts, age_act_pars)

        # Apply interventions
        for i,intervention in enumerate(self.interventions):
            intervention(self) # If it's a function, call it directly

        # Assign sus_imm values, i.e. the protection against infection based on prior immune history
        if self['use_waning']:
            inds = hpu.true(people.peak_imm.sum(axis=0)).astype(hpd.default_int)
            if len(inds):
                ss = people.t_imm_event[:, inds].shape
                t_since_boost = (t - people.t_imm_event[:,inds]).ravel()
                current_imm = imm_kin_pars[t_since_boost].reshape(ss) # Get people's current level of immunity
                people.nab_imm[:,inds] = current_imm*people.peak_imm[:,inds] # Set immunity relative to peak
                    # return imm
        else:
            people.nab_imm[:] = people.peak_imm
        hpimm.check_immunity(people)

        # Shorten more variables
        gen_betas = np.array([g['rel_beta'] * beta for g in gen_pars.values()], dtype=hpd.default_float)
        sus_imm = people.sus_imm
        hiv_rel_sus = np.ones(len(people), dtype=hpd.default_float)
        hiv_inds = hpu.true(people.hiv)
        immune_compromise = 1 - people.art_adherence[hiv_inds]
        mod = immune_compromise * self['hiv_pars']['rel_sus']
        mod[mod < 1] = 1
        hiv_rel_sus[hiv_inds] *= mod

        inf = people.infectious.copy() # calculate transmission based on infectiousness at start of timestep i.e. someone infected in one layer cannot transmit the infection via a different layer in the same timestep

        # Loop over layers
        for lkey, layer in people.contacts.items():

            sus = people.susceptible.copy() # for each layer, update who's still susceptible

            # Shorten variables
            f = layer['f']
            m = layer['m']
            acts = layer['acts'] * dt
            frac_acts, whole_acts = np.modf(acts)
            whole_acts = whole_acts.astype(hpd.default_int)
            effective_condoms = hpd.default_float(condoms[lkey] * eff_condoms)

            # Compute transmissions by genotype
            for g in range(ng):

                f_source_inds = (inf[g][f] & sus[g][m]).nonzero()[0]  # get female sources where female partner is infectious with genotype and male partner is susceptible to that genotype
                m_source_inds = (inf[g][m] & sus[g][f]).nonzero()[0]  # get male sources where the male partner is infectious with genotype and the female partner is susceptible to that genotype

                foi_frac = 1 - frac_acts * gen_betas[g] * trans[:, None] * (1 - effective_condoms)  # Probability of not getting infected from any fractional acts
                foi_whole = (1 - gen_betas[g] * trans[:, None] * (1 - effective_condoms)) ** whole_acts  # Probability of not getting infected from whole acts
                foi = (1 - (foi_whole * foi_frac)).astype(hpd.default_float)

                discordant_pairs = [[f_source_inds, f[f_source_inds], m[f_source_inds], foi[0,:]],
                                    [m_source_inds, m[m_source_inds], f[m_source_inds], foi[1,:]]]

                # Compute transmissibility for each partnership
                for pship_inds, sources, targets, this_foi in discordant_pairs:
                    betas = this_foi[pship_inds] * (1. - sus_imm[g,targets]) * hiv_rel_sus[targets] # Pull out the transmissibility associated with this partnership
                    transmissions = (np.random.random(len(betas)) < betas).nonzero()[0] # Apply probabilities to determine partnerships in which transmission occurred
                    target_inds   = targets[transmissions] # Extract indices of those who got infected
                    target_inds, unique_inds = np.unique(target_inds, return_index=True)  # Due to multiple partnerships, some people will be counted twice; remove them
                    people.infect(inds=target_inds, g=g, layer=lkey)  # Infect people

        # Determine if there are any reactivated infections on this timestep
        for g in range(ng):
            latent_inds = hpu.true(people.latent[g,:])
            if len(latent_inds):
                reactivation_probs = np.full_like(latent_inds, self['hpv_reactivation'] * dt, dtype=hpd.default_float)

                if self['model_hiv']:
                    # determine if any of these inds have HIV and adjust their probs
                    hiv_latent_inds = latent_inds[hpu.true(people.hiv[latent_inds])]
                    if len(hiv_latent_inds):
                        immune_compromise = 1 - people.art_adherence[hiv_latent_inds]
                        mod = immune_compromise * self['hiv_pars']['reactivation_prob']
                        mod[mod < 1] = 1
                        reactivation_probs[hpu.true(people.hiv[latent_inds])] *= mod
                is_reactivated = hpu.binomial_arr(reactivation_probs)
                reactivated_inds = latent_inds[is_reactivated]
                people.infect(inds=reactivated_inds, g=g, layer='reactivation')

        # Index for results
        idx = int(t / self.resfreq)

        # Update counts for this time step: flows
        for key,count in people.flows.items():
            self.results[key][idx] += count
        for key,count in people.demographic_flows.items():
            self.results[key][idx] += count
        for key,count in people.genotype_flows.items():
            flow_ind = [flow.name for flow in hpd.flows].index(key)
            if hpd.flows[flow_ind].by_genotype:
                for genotype in range(ng):
                    self.results[key+'_by_genotype'][genotype][idx] += count[genotype]
        for key,count in people.sex_flows.items():
            for sex in range(2):
                self.results[key][sex][idx] += count[sex]
        for key,count in people.age_flows.items():
            self.results[key+'_by_age'][:,idx] += count

        # Make stock updates every nth step, where n is the frequency of result output
        if t % self.resfreq == self.resfreq-1:

            # Number infectious/susceptible by age, for prevalence calculations
            infinds = hpu.true(people['infectious'])
            self.results[f'n_infectious_by_age'][:, idx] = np.histogram(people.age[infinds], bins=people.age_bins, weights=people.scale[infinds])[0]
            susinds = hpu.true(people['susceptible'])
            self.results[f'n_susceptible_by_age'][:, idx] = np.histogram(people.age[susinds], bins=people.age_bins, weights=people.scale[susinds])[0]
            transformedinds = hpu.true(people['transformed'])
            self.results[f'n_transformed_by_age'][:, idx] = np.histogram(people.age[transformedinds], bins=people.age_bins, weights=people.scale[transformedinds])[0]
            precininds = hpu.true(people['precin'])
            self.results[f'n_precin_by_age'][:, idx] = np.histogram(people.age[precininds], bins=people.age_bins, weights=people.scale[precininds])[0]
            cin1inds = hpu.true(people['cin1'])
            self.results[f'n_cin1_by_age'][:, idx] = np.histogram(people.age[cin1inds], bins=people.age_bins, weights=people.scale[cin1inds])[0]
            cin2inds = hpu.true(people['cin2'])
            self.results[f'n_cin2_by_age'][:, idx] = np.histogram(people.age[cin2inds], bins=people.age_bins, weights=people.scale[cin2inds])[0]
            cin3inds = hpu.true(people['cin3'])
            self.results[f'n_cin3_by_age'][:, idx] = np.histogram(people.age[cin3inds], bins=people.age_bins, weights=people.scale[cin3inds])[0]
            # Create total stocks
            for key in hpd.total_stock_keys:

                # Stocks by genotype
                for g in range(ng):
                    self.results[f'n_{key}_by_genotype'][g, idx] = people.count_by_genotype(key, g)

                # Total stocks
                if key not in ['susceptible']:
                    # For n_infectious etc, we get the total number where this state is true for at least one genotype
                    self.results[f'n_{key}'][idx] = people.count_any(key)
                elif key == 'susceptible':
                    # For n_total_susceptible, we get the total number of infections that could theoretically happen in the population, which can be greater than the population size
                    self.results[f'n_{key}'][idx] = people.count(key)

            # Create stocks of interventions
            for key in [state.name for state in hpd.PeopleMeta.intv_states]:
                self.results[f'n_{key}'][idx] = people.count(key)

            # Count total hiv infections
            self.results['n_hiv'][idx] = people.count('hiv')

            # Update cancers and cancers by age
            cases_by_age = self.results['cancers_by_age'][:, idx]
            inds = people.alive * (self.people.sex==0) * ~people.cancerous.any(axis=0)
            vals = self.people.age[inds]
            bins = self.pars['standard_pop'][0,]
            weights = people.scale[inds]
            denom = np.histogram(vals, bins, weights=weights)[0]
            age_specific_incidence = sc.safedivide(cases_by_age, denom)*100e3
            standard_pop = self.pars['standard_pop'][1, :-1]
            self.results['asr_cancer_incidence'][idx] = np.dot(age_specific_incidence,standard_pop)

            # Save number alive
            alive_inds = hpu.true(people.alive)
            alive_female_inds = hpu.true(people.alive*people.is_female)
            self.results['n_alive'][idx] = people.scale_flows(alive_inds)
            self.results['n_alive_by_sex'][0,idx] = people.scale_flows((people.alive*people.is_female).nonzero()[0])
            self.results['n_alive_by_sex'][1,idx] = people.scale_flows((people.alive*people.is_male).nonzero()[0])
            self.results['n_alive_by_age'][:,idx] = np.histogram(people.age[alive_inds], bins=people.age_bins, weights=people.scale[alive_inds])[0]
            self.results['n_females_alive_by_age'][:,idx] = np.histogram(people.age[alive_female_inds], bins=people.age_bins, weights=people.scale[alive_female_inds])[0]

        # Apply analyzers
        for i,analyzer in enumerate(self.analyzers):
            analyzer(self)

        # Tidy up
        self.t += 1
        if self.t == self.npts:
            self.complete = True

        return


    def run(self, do_plot=False, until=None, restore_pars=True, reset_seed=True, verbose=None):
        ''' Run the model once '''
        # Initialization steps -- start the timer, initialize the sim and the seed, and check that the sim hasn't been run
        T = sc.timer()

        if not self.initialized:
            self.initialize()
            self._orig_pars = sc.dcp(self.pars) # Create a copy of the parameters, to restore after the run, in case they are dynamically modified

        if verbose is None:
            verbose = self['verbose']

        if reset_seed:
            # Reset the RNG. The primary use case (and why it defaults to True) is to ensure that
            #
            # >>> sim0.initialize()
            # >>> sim0.run()
            # >>> sim1.initialize()
            # >>> sim1.run()
            #
            # produces the same output as
            #
            # >>> sim0.initialize()
            # >>> sim1.initialize()
            # >>> sim0.run()
            # >>> sim1.run()
            #
            # The seed is offset by 1 to avoid drawing the same random numbers as those used for population generation, otherwise
            # the first set of random numbers in the model (e.g., deaths) will be correlated with the first set of random numbers
            # drawn in population generation (e.g., sex)
            hpu.set_seed(self['rand_seed']+1)

        # Check for AlreadyRun errors
        errormsg = None
        until = self.npts if until is None else self.get_t(until)
        if until > self.npts:
            errormsg = f'Requested to run until t={until} but the simulation end is t={self.npts}'
        if self.t >= until: # NB. At the start, self.t is None so this check must occur after initialization
            errormsg = f'Simulation is currently at t={self.t}, requested to run until t={until} which has already been reached'
        if self.complete:
            errormsg = 'Simulation is already complete (call sim.initialize() to re-run)'
        if self.people.t not in [self.t, self.t-1]: # Depending on how the sim stopped, either of these states are possible
            errormsg = f'The simulation has been run independently from the people (t={self.t}, people.t={self.people.t}): if this is intentional, manually set sim.people.t = sim.t. Remember to save the people object before running the sim.'
        if errormsg:
            raise AlreadyRunError(errormsg)

        # Main simulation loop
        while self.t < until:

            # Check if we were asked to stop
            elapsed = T.toc(output=True)
            if self['timelimit'] and elapsed > self['timelimit']:
                sc.printv(f"Time limit ({self['timelimit']} s) exceeded; call sim.finalize() to compute results if desired", 1, verbose)
                return
            elif self['stopping_func'] and self['stopping_func'](self):
                sc.printv("Stopping function terminated the simulation; call sim.finalize() to compute results if desired", 1, verbose)
                return

            # Print progress
            if verbose:
                simlabel = f'"{self.label}": ' if self.label else ''
                string = f'  Running {simlabel}{self.yearvec[self.t]:0.1f} ({self.t:2.0f}/{self.npts}) ({elapsed:0.2f} s) '
                if verbose >= 2:
                    sc.heading(string)
                elif verbose>0:
                    if not (self.t % int(1.0/verbose)):
                        sc.progressbar(self.t+1, self.npts, label=string, length=20, newline=True)

            # Do the heavy lifting -- actually run the model!
            self.step()

        # If simulation reached the end, finalize the results
        if self.complete:
            self.finalize(verbose=verbose, restore_pars=restore_pars)
            sc.printv(f'Run finished after {elapsed:0.2f} s.\n', 1, verbose)

        return self


    def finalize(self, verbose=None, restore_pars=True):
        ''' Compute final results '''

        if self.results_ready:
            # Because the results are rescaled in-place, finalizing the sim cannot be run more than once or
            # otherwise the scale factor will be applied multiple times
            raise AlreadyRunError('Simulation has already been finalized')

        # Finalize analyzers and interventions
        self.finalize_analyzers()
        # self.finalize_interventions()

        # Final settings
        self.results_ready = True # Set this first so self.summary() knows to print the results
        self.t -= 1 # During the run, this keeps track of the next step; restore this be the final day of the sim

        # Perform calculations on results
        self.compute_results(verbose=verbose) # Calculate the rest of the results
        self.results = sc.objdict(self.results) # Convert results to a odicts/objdict to allow e.g. sim.results.diagnoses

        # Optionally print summary output
        if verbose: # Verbose is any non-zero value
            if verbose>0: # Verbose is any positive number
                self.summarize() # Print medium-length summary of the sim
            else:
                self.brief() # Print brief summary of the sim

        return


    def compute_results(self, verbose=None):
        ''' Perform final calculations on the results '''
        self.compute_states()
        self.compute_summary()
        return


    def compute_states(self):
        '''
        Compute prevalence, incidence, and other states.
        '''
        res = self.results

        # Compute HPV incidence and prevalence
        def safedivide(num,denom):
            ''' Define a variation on sc.safedivide that respects shape of numerator '''
            answer = np.zeros_like(num)
            fill_inds = (denom!=0).nonzero()
            if len(num.shape)==len(denom.shape):
                answer[fill_inds] = num[fill_inds] / denom[fill_inds]
            else:
                answer[:, fill_inds] = num[:, fill_inds] / denom[fill_inds]
            return answer
        ng = self.pars['n_genotypes']
        self.results['hpv_incidence'][:]                = sc.safedivide(res['infections'][:], ng*res['n_susceptible'][:])
        self.results['hpv_incidence_by_genotype'][:]    = safedivide(res['infections_by_genotype'][:], res['n_susceptible_by_genotype'][:])
        self.results['hpv_incidence_by_age'][:]         = safedivide(res['infections_by_age'][:], ng*res['n_susceptible_by_age'][:])
        self.results['hpv_prevalence'][:]               = sc.safedivide(res['n_infectious'][:], ng*res['n_alive'][:])
        self.results['hpv_prevalence_by_genotype'][:]   = safedivide(res['n_infectious_by_genotype'][:], res['n_alive'][:])
        self.results['hpv_prevalence_by_age'][:]        = safedivide(res['n_infectious_by_age'][:], ng*res['n_alive_by_age'][:])
        self.results['hiv_incidence'][:]                = sc.safedivide(res['hiv_infections'][:], (res['n_alive'][:]-res['n_hiv'][:]))
        self.results['hiv_prevalence'][:]               = sc.safedivide(res['n_hiv'][:], res['n_alive'][:])

        alive_females = res['n_alive_by_sex'][0,:]

        self.results['precin_prevalence'][:] = sc.safedivide(res['n_precin'][:], ng * alive_females)
        self.results['precin_prevalence_by_genotype'][:] = safedivide(res['n_precin_by_genotype'][:], alive_females)
        self.results['precin_prevalence_by_age'][:] = safedivide(res['n_precin_by_age'][:],
                                                               ng * res['n_females_alive_by_age'][:])
        self.results['cin1_prevalence'][:] = sc.safedivide(res['n_cin1'][:], ng*alive_females)
        self.results['cin1_prevalence_by_genotype'][:] = safedivide(res['n_cin1_by_genotype'][:], alive_females)
        self.results['cin1_prevalence_by_age'][:] = safedivide(res['n_cin1_by_age'][:],
                                                               ng*res['n_females_alive_by_age'][:])
        self.results['cin2_prevalence'][:] = sc.safedivide(res['n_cin2'][:], ng*alive_females)
        self.results['cin2_prevalence_by_genotype'][:] = safedivide(res['n_cin2_by_genotype'][:], alive_females)
        self.results['cin2_prevalence_by_age'][:] = safedivide(res['n_cin2_by_age'][:],
                                                               ng*res['n_females_alive_by_age'][:])
        self.results['cin3_prevalence'][:] = sc.safedivide(res['n_cin3'][:], ng*alive_females)
        self.results['cin3_prevalence_by_genotype'][:] = safedivide(res['n_cin3_by_genotype'][:], alive_females)
        self.results['cin3_prevalence_by_age'][:] = safedivide(res['n_cin3_by_age'][:],
                                                               ng*res['n_females_alive_by_age'][:])
        # Compute CIN and cancer incidence.
        at_risk_females = alive_females - res['n_cancerous'][:]
        scale_factor = 1e5  # Cancer and CIN incidence are displayed as rates per 100k women
        demoninator = at_risk_females / scale_factor
        self.results['cancer_incidence'][:]             = res['cancers'][:] / demoninator
        self.results['cancer_incidence_by_genotype'][:] = res['cancers_by_genotype'][:] / demoninator
        self.results['cancer_incidence_by_age'][:]      = sc.safedivide(res['cancers'][:], res['n_females_alive_by_age'][:]/scale_factor)

        # Compute cancer mortality. Denominator is all women alive
        denominator = alive_females/scale_factor
        self.results['cancer_mortality'][:]         = res['cancer_deaths'][:]/denominator

        # Compute HPV type distribution by cytology
        for which in hpd.type_dist_keys:
            by_type = res[f'n_{which}_by_genotype'][:]
            totals = by_type.sum(axis=0)
            inds_to_fill = totals > 0
            res[which + '_genotype_dist'][:, inds_to_fill] = by_type[:, inds_to_fill] / totals[inds_to_fill]

        # Demographic results
        self.results['cdr'][:]  = self.results['other_deaths'][:] / (self.results['n_alive'][:])
        self.results['cbr'][:]  = self.results['births'][:] / (self.results['n_alive'][:])

        # Vaccination results
        self.results['cum_vaccinated'][:] = np.cumsum(self.results['new_vaccinated'][:], axis=0)
        self.results['cum_total_vaccinated'][:] = np.cumsum(self.results['new_total_vaccinated'][:])
        self.results['cum_doses'][:] = np.cumsum(self.results['new_doses'][:])

        # Therapeutic vaccination results
        self.results['cum_tx_vaccinated'][:] = np.cumsum(self.results['new_tx_vaccinated'][:], axis=0)
        self.results['cum_txvx_doses'][:] = np.cumsum(self.results['new_txvx_doses'][:])

        # Screen & treat results
        self.results['cum_screens'][:] = np.cumsum(self.results['new_screens'][:], axis=0)
        self.results['cum_screened'][:] = np.cumsum(self.results['new_screened'][:], axis=0)
        self.results['cum_cin_treatments'][:] = np.cumsum(self.results['new_cin_treatments'][:], axis=0)
        self.results['cum_cin_treated'][:] = np.cumsum(self.results['new_cin_treated'][:], axis=0)
        self.results['cum_cancer_treatments'][:] = np.cumsum(self.results['new_cancer_treatments'][:], axis=0)
        self.results['cum_cancer_treated'][:] = np.cumsum(self.results['new_cancer_treatments'][:], axis=0)

        return


    def compute_summary(self, t=None, update=True, output=False, require_run=False):
        '''
        Compute the summary dict and string for the sim. Used internally; see
        sim.summarize() for the user version.

        Args:
            t (int/str): day or date to compute summary for (by default, the last point)
            update (bool): whether to update the stored sim.summary
            output (bool): whether to return the summary
            require_run (bool): whether to raise an exception if simulations have not been run yet
        '''
        if t is None:
            t = -1

        # Compute the summary
        if require_run and not self.results_ready:
            errormsg = 'Simulation not yet run'
            raise RuntimeError(errormsg)

        summary = sc.objdict()
        for key in self.result_keys('total'):
            summary[key] = self.results[key][t]

        # Update the stored state
        if update:
            self.summary = summary

        # Optionally return
        if output:
            return summary
        else:
            return


    def summarize(self, full=False, t=None, sep=None, output=False):
        '''
        Print a medium-length summary of the simulation, drawing from the last time
        point in the simulation by default. Called by default at the end of a sim run.
        point in the simulation by default. Called by default at the end of a sim run.
        See also sim.disp() (detailed output) and sim.brief() (short output).

        Args:
            full   (bool):    whether or not to print all results (by default, only cumulative)
            t      (int/str): day or date to compute summary for (by default, the last point)
            sep    (str):     thousands separator (default ',')
            output (bool):    whether to return the summary instead of printing it

        **Examples**::

            sim = hpv.Sim(label='Example sim', verbose=0) # Set to run silently
            sim.run() # Run the sim
            sim.summarize() # Print medium-length summary of the sim
            sim.summarize(t=24, full=True) # Print a "slice" of all sim results on day 24
        '''
        # Compute the summary
        summary = self.compute_summary(t=t, update=False, output=True)

        # Construct the output string
        if sep is None: sep = hpo.sep # Default separator
        labelstr = f' "{self.label}"' if self.label else ''
        string = f'Simulation{labelstr} summary:\n'
        for key in self.result_keys('total'):
            val = summary[key]
            printval = f'   {val:10,.0f} '
            label = self.results[key].name.lower().replace(',', sep)
            if 'incidence' in key or 'prevalence' in key:
                if key in ['hpv_prevalence', 'hpv_incidence']:
                    printval = f'   {val*100:10.2f} '
                    label += ' (/100)'
                else:
                    label += ' (/100,000)'
            string += printval + label + '\n'

        # Print or return string
        if not output:
            print(string)
        else:
            return string


    def plot(self, *args, **kwargs):
        '''
        Plot the outputs of the model
        '''
        fig = hpplt.plot_sim(sim=self, *args, **kwargs)
        return fig


    def compute_fit(self):
        '''
        Compute fit between model and data.
        '''
        return self.fit


class AlreadyRunError(RuntimeError):
    '''
    This error is raised if a simulation is run in such a way that no timesteps
    will be taken. This error is a distinct type so that it can be safely caught
    and ignored if required, but it is anticipated that most of the time, calling
    :py:func:`Sim.run` and not taking any timesteps, would be an inadvertent error.
    '''
    pass


def diff_sims(sim1, sim2, skip_key_diffs=False, skip=None, output=False, die=False):
    '''
    Compute the difference of the summaries of two simulations, and print any
    values which differ.

    Args:
        sim1 (sim/dict): either a simulation object or the sim.summary dictionary
        sim2 (sim/dict): ditto
        skip_key_diffs (bool): whether to skip keys that don't match between sims
        skip (list): a list of values to skip
        output (bool): whether to return the output as a string (otherwise print)
        die (bool): whether to raise an exception if the sims don't match
        require_run (bool): require that the simulations have been run

    **Example**::

        s1 = hpv.Sim(rand_seed=1).run()
        s2 = hpv.Sim(rand_seed=2).run()
        hpv.diff_sims(s1, s2)
    '''

    if isinstance(sim1, Sim):
        sim1 = sim1.compute_summary(update=False, output=True, require_run=True)
    if isinstance(sim2, Sim):
        sim2 = sim2.compute_summary(update=False, output=True, require_run=True)
    for sim in [sim1, sim2]:
        if not isinstance(sim, dict): # pragma: no cover
            errormsg = f'Cannot compare object of type {type(sim)}, must be a sim or a sim.summary dict'
            raise TypeError(errormsg)

    # Compare keys
    keymatchmsg = ''
    sim1_keys = set(sim1.keys())
    sim2_keys = set(sim2.keys())
    if sim1_keys != sim2_keys and not skip_key_diffs: # pragma: no cover
        keymatchmsg = "Keys don't match!\n"
        missing = list(sim1_keys - sim2_keys)
        extra   = list(sim2_keys - sim1_keys)
        if missing:
            keymatchmsg += f'  Missing sim1 keys: {missing}\ns'
        if extra:
            keymatchmsg += f'  Extra sim2 keys: {extra}\n'

    # Compare values
    valmatchmsg = ''
    mismatches = {}
    skip = sc.tolist(skip)
    for key in sim2.keys(): # To ensure order
        if key in sim1_keys and key not in skip: # If a key is missing, don't count it as a mismatch
            sim1_val = sim1[key] if key in sim1 else 'not present'
            sim2_val = sim2[key] if key in sim2 else 'not present'
            if not np.isclose(sim1_val, sim2_val, equal_nan=True):
                mismatches[key] = {'sim1': sim1_val, 'sim2': sim2_val}

    if len(mismatches):
        valmatchmsg = '\nThe following values differ between the two simulations:\n'
        df = pd.DataFrame.from_dict(mismatches).transpose()
        diff   = []
        ratio  = []
        change = []
        small_change = 1e-3 # Define a small change, e.g. a rounding error
        for mdict in mismatches.values():
            old = mdict['sim1']
            new = mdict['sim2']
            numeric = sc.isnumber(sim1_val) and sc.isnumber(sim2_val)
            if numeric and old>0:
                this_diff  = new - old
                this_ratio = new/old
                abs_ratio  = max(this_ratio, 1.0/this_ratio)

                # Set the character to use
                if abs_ratio<small_change:
                    change_char = '≈'
                elif new > old:
                    change_char = '↑'
                elif new < old:
                    change_char = '↓'
                else:
                    errormsg = f'Could not determine relationship between sim1={old} and sim2={new}'
                    raise ValueError(errormsg)

                # Set how many repeats it should have
                repeats = 1
                if abs_ratio >= 1.1:
                    repeats = 2
                if abs_ratio >= 2:
                    repeats = 3
                if abs_ratio >= 10:
                    repeats = 4

                this_change = change_char*repeats
            else: # pragma: no cover
                this_diff   = np.nan
                this_ratio  = np.nan
                this_change = 'N/A'

            diff.append(this_diff)
            ratio.append(this_ratio)
            change.append(this_change)

        df['diff'] = diff
        df['ratio'] = ratio
        for col in ['sim1', 'sim2', 'diff', 'ratio']:
            df[col] = df[col].round(decimals=3)
        df['change'] = change
        valmatchmsg += str(df)

    # Raise an error if mismatches were found
    mismatchmsg = keymatchmsg + valmatchmsg
    if mismatchmsg: # pragma: no cover
        if die:
            raise ValueError(mismatchmsg)
        elif output:
            return mismatchmsg
        else:
            print(mismatchmsg)
    else:
        if not output:
            print('Sims match')
    return
